#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
import warnings

from pgmpy.readwrite import PomdpXReader, PomdpXWriter
from pgmpy.extern import six

try:
    from lxml import etree
except ImportError:
    try:
        # Python 2.5
        import xml.etree.cElementTree as etree
        six.print_("running with cElementTree on Python 2.5+")
    except ImportError:
        try:
            import xml.etree.ElementTree as etree
            print("running with ElementTree on Python 2.5+")
        except ImportError:
            warnings.warn("Failed to import ElementTree from any known place")


class TestPomdpXReaderString(unittest.TestCase):
    def setUp(self):
        string = """<pomdpx version="1.0" id="rockSample"
      xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
      xsi:noNamespaceSchemaLocation="pomdpx.xsd">
         <Description>RockSample problem for map size 1 x 3.
           Rock is at 0, Rover’s initial position is at 1.
           Exit is at 2.
         </Description>
         <Discount>0.95</Discount>
         <Variable>
              <StateVar vnamePrev="rover_0" vnameCurr="rover_1"
                fullyObs="true">
                  <NumValues>3</NumValues>
              </StateVar>
              <StateVar vnamePrev="rock_0" vnameCurr="rock_1">
                  <ValueEnum>good bad</ValueEnum>
              </StateVar>
              <ObsVar vname="obs_sensor">
                  <ValueEnum>ogood obad</ValueEnum>
              </ObsVar>
              <ActionVar vname="action_rover">
                  <ValueEnum>amw ame ac as</ValueEnum>
              </ActionVar>
              <RewardVar vname="reward_rover" />
         </Variable>
         <InitialStateBelief>
              <CondProb>
                  <Var>rover_0</Var>
                  <Parent>null</Parent>
                  <Parameter type="TBL">
                        <Entry>
                            <Instance> - </Instance>
                            <ProbTable>0.0 1.0 0.0</ProbTable>
                        </Entry>
              </Parameter>
         </CondProb>
         <CondProb>
              <Var>rock_0</Var>
              <Parent>null</Parent>
              <Parameter type="TBL">
                  <Entry>
                      <Instance>-</Instance>
                      <ProbTable>uniform</ProbTable>
                  </Entry>
              </Parameter>
         </CondProb>
      </InitialStateBelief>
      <StateTransitionFunction>
          <CondProb>
              <Var>rover_1</Var>
              <Parent>action_rover rover_0</Parent>
              <Parameter type="TBL">
                  <Entry>
                      <Instance>amw s0 s2</Instance>
                      <ProbTable>1.0</ProbTable>
                  </Entry>
                  <Entry>
                      <Instance>amw s1 s0</Instance>
                      <ProbTable>1.0</ProbTable>
                  </Entry>
                  <Entry>
                      <Instance>ame s0 s1</Instance>
                      <ProbTable>1.0</ProbTable>
                  </Entry>
                  <Entry>
                      <Instance>ame s1 s2</Instance>
                      <ProbTable>1.0</ProbTable>
                  </Entry>
                  <Entry>
                      <Instance>ac s0 s0</Instance>
                      <ProbTable>1.0</ProbTable>
                  </Entry>
                  <Entry>
                      <Instance>ac s1 s1</Instance>
                      <ProbTable>1.0</ProbTable>
                  </Entry>
                  <Entry>
                      <Instance>as s0 s0</Instance>
                      <ProbTable>1.0</ProbTable>
                  </Entry>
                  <Entry>
                      <Instance>as s1 s2</Instance>
                      <ProbTable>1.0</ProbTable>
                  </Entry>
                  <Entry>
                      <Instance>* s2 s2</Instance>
                      <ProbTable>1.0</ProbTable>
                  </Entry>
           </Parameter>
       </CondProb>
       <CondProb>
           <Var>rock_1</Var>
           <Parent>action_rover rover_0 rock_0</Parent>
           <Parameter>
               <Entry>
                   <Instance>amw * - - </Instance>
                   <ProbTable>1.0 0.0 0.0 1.0</ProbTable>
               </Entry>
               <Entry>
                   <Instance>ame * - - </Instance>
                   <ProbTable>identity</ProbTable>
               </Entry>
               <Entry>
                   <Instance>ac * - - </Instance>
                   <ProbTable>identity</ProbTable>
               </Entry>
               <Entry>
                   <Instance>as * - - </Instance>
                   <ProbTable>identity</ProbTable>
               </Entry>
               <Entry>
                   <Instance>as s0 * - </Instance>
                   <ProbTable>0.0 1.0</ProbTable>
               </Entry>
           </Parameter>
       </CondProb>
   </StateTransitionFunction>
   <ObsFunction>
       <CondProb>
           <Var>obs_sensor</Var>
           <Parent>action_rover rover_1 rock_1</Parent>
           <Parameter type="TBL">
               <Entry>
                   <Instance>amw * * - </Instance>
                   <ProbTable>1.0 0.0</ProbTable>
               </Entry>
               <Entry>
                   <Instance>ame * * - </Instance>
                   <ProbTable>1.0 0.0</ProbTable>
               </Entry>
               <Entry>
                   <Instance>as * * - </Instance>
                   <ProbTable>1.0 0.0</ProbTable>
               </Entry>
               <Entry>
                   <Instance>ac s0 - - </Instance>
                   <ProbTable>1.0 0.0 0.0 1.0</ProbTable>
               </Entry>
               <Entry>
                   <Instance>ac s1 - - </Instance>
                   <ProbTable>0.8 0.2 0.2 0.8</ProbTable>
               </Entry>
                <Entry>
                     <Instance>ac s2 * - </Instance>
                     <ProbTable>1.0 0.0</ProbTable>
                </Entry>
            </Parameter>
        </CondProb>
    </ObsFunction>
    <RewardFunction>
        <Func>
            <Var>reward_rover</Var>
            <Parent>action_rover rover_0 rock_0</Parent>
            <Parameter type="TBL">
                <Entry>
                     <Instance>ame s1 *</Instance>
                     <ValueTable>10</ValueTable>
                </Entry>
                <Entry>
                     <Instance>amw s0 *</Instance>
                     <ValueTable>-100</ValueTable>
                </Entry>
                <Entry>
                     <Instance>as s1 *</Instance>
                     <ValueTable>-100</ValueTable>
                </Entry>
                <Entry>
                     <Instance>as s0 good</Instance>
                     <ValueTable>10</ValueTable>
                </Entry>
                <Entry>
                     <Instance>as s0 bad</Instance>
                     <ValueTable>-10</ValueTable>
                </Entry>
            </Parameter>
        </Func>
    </RewardFunction>
 </pomdpx>
 """
        self.reader_string = PomdpXReader(string=string)
        self.reader_file = PomdpXReader(path=six.StringIO(string))

    def test_get_variables(self):
        var_expected = {'StateVar': [
                        {'vnamePrev': 'rover_0',
                         'vnameCurr': 'rover_1',
                         'ValueEnum': ['s0', 's1', 's2'],
                         'fullyObs': True},
                        {'vnamePrev': 'rock_0',
                         'vnameCurr': 'rock_1',
                         'fullyObs': False,
                         'ValueEnum': ['good', 'bad']}],
                        'ObsVar': [{'vname': 'obs_sensor',
                                    'ValueEnum': ['ogood', 'obad']}],
                        'RewardVar': [{'vname': 'reward_rover'}],
                        'ActionVar': [{'vname': 'action_rover',
                                       'ValueEnum': ['amw', 'ame',
                                                     'ac', 'as']}]
                        }
        self.maxDiff = None
        self.assertEqual(self.reader_string.get_variables(), var_expected)
        self.assertEqual(self.reader_file.get_variables(), var_expected)

    def test_get_initial_belief_system(self):
        belief_expected = [{'Var': 'rover_0',
                            'Parent': ['null'],
                            'Type': 'TBL',
                            'Parameter': [{'Instance': ['-'],
                                           'ProbTable': ['0.0', '1.0', '0.0']}]
                            },
                           {'Var': 'rock_0',
                            'Parent': ['null'],
                            'Type': 'TBL',
                            'Parameter': [{'Instance': ['-'],
                                           'ProbTable': ['uniform']}]
                            }]
        self.maxDiff = None
        self.assertEqual(self.reader_string.get_initial_beliefs(), belief_expected)
        self.assertEqual(self.reader_file.get_initial_beliefs(), belief_expected)

    def test_get_state_transition_function(self):
        state_transition_function_expected = \
            [{'Var': 'rover_1',
              'Parent': ['action_rover', 'rover_0'],
              'Type': 'TBL',
              'Parameter': [{'Instance': ['amw', 's0', 's2'],
                             'ProbTable': ['1.0']},
                            {'Instance': ['amw', 's1', 's0'],
                             'ProbTable': ['1.0']},
                            {'Instance': ['ame', 's0', 's1'],
                             'ProbTable': ['1.0']},
                            {'Instance': ['ame', 's1', 's2'],
                             'ProbTable': ['1.0']},
                            {'Instance': ['ac', 's0', 's0'],
                             'ProbTable': ['1.0']},
                            {'Instance': ['ac', 's1', 's1'],
                             'ProbTable': ['1.0']},
                            {'Instance': ['as', 's0', 's0'],
                             'ProbTable': ['1.0']},
                            {'Instance': ['as', 's1', 's2'],
                             'ProbTable': ['1.0']},
                            {'Instance': ['*', 's2', 's2'],
                             'ProbTable': ['1.0']}]},
             {'Var': 'rock_1',
              'Parent': ['action_rover', 'rover_0', 'rock_0'],
              'Type': 'TBL',
              'Parameter': [{'Instance': ['amw', '*', '-', '-'],
                             'ProbTable': ['1.0', '0.0', '0.0', '1.0']},
                            {'Instance': ['ame', '*', '-', '-'],
                             'ProbTable': ['identity']},
                            {'Instance': ['ac', '*', '-', '-'],
                             'ProbTable': ['identity']},
                            {'Instance': ['as', '*', '-', '-'],
                             'ProbTable': ['identity']},
                            {'Instance': ['as', 's0', '*', '-'],
                             'ProbTable': ['0.0', '1.0']},
                            ]}]
        self.maxDiff = None
        self.assertEqual(self.reader_string.get_state_transition_function(),
                         state_transition_function_expected)
        self.assertEqual(self.reader_file.get_state_transition_function(),
                         state_transition_function_expected)

    def test_obs_function(self):
        obs_function_expected = \
            [{'Var': 'obs_sensor',
              'Parent': ['action_rover', 'rover_1', 'rock_1'],
              'Type': 'TBL',
              'Parameter': [{'Instance': ['amw', '*', '*', '-'],
                             'ProbTable': ['1.0', '0.0']},
                            {'Instance': ['ame', '*', '*', '-'],
                             'ProbTable': ['1.0', '0.0']},
                            {'Instance': ['as', '*', '*', '-'],
                             'ProbTable': ['1.0', '0.0']},
                            {'Instance': ['ac', 's0', '-', '-'],
                             'ProbTable': ['1.0', '0.0', '0.0', '1.0']},
                            {'Instance': ['ac', 's1', '-', '-'],
                             'ProbTable': ['0.8', '0.2', '0.2', '0.8']},
                            {'Instance': ['ac', 's2', '*', '-'],
                             'ProbTable': ['1.0', '0.0']}]}]
        self.maxDiff = None
        self.assertEqual(self.reader_string.get_obs_function(), obs_function_expected)
        self.assertEqual(self.reader_file.get_obs_function(), obs_function_expected)

    def test_reward_function(self):
        reward_function_expected = \
            [{'Var': 'reward_rover',
              'Parent': ['action_rover', 'rover_0', 'rock_0'],
              'Type': 'TBL',
              'Parameter': [{'Instance': ['ame', 's1', '*'],
                             'ValueTable': ['10']},
                            {'Instance': ['amw', 's0', '*'],
                             'ValueTable': ['-100']},
                            {'Instance': ['as', 's1', '*'],
                             'ValueTable': ['-100']},
                            {'Instance': ['as', 's0', 'good'],
                             'ValueTable': ['10']},
                            {'Instance': ['as', 's0', 'bad'],
                             'ValueTable': ['-10']}]}]
        self.maxDiff = None
        self.assertEqual(self.reader_string.get_reward_function(),
                         reward_function_expected)
        self.assertEqual(self.reader_file.get_reward_function(),
                         reward_function_expected)

    def test_get_parameter_dd(self):
        string = """
        <pomdpx version="1.0" id="rockSample"
      xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
      xsi:noNamespaceSchemaLocation="pomdpx.xsd">
         <Description>RockSample problem for map size 1 x 3.
           Rock is at 0, Rover’s initial position is at 1.
           Exit is at 2.
         </Description>
         <Discount>0.95</Discount>
        <InitialStateBelief>
        <CondProb>
         <Var>rover_0</Var>
  <Parent>null</Parent>
   <Parameter type = "DD">
      <DAG>
         <Node var = "rover_0">
             <Edge val="s0"><Terminal>0.0</Terminal></Edge>
              <Edge val="s1">
                 <Node var = "rock_0">
                     <Edge val = "good">
                       <Terminal>0.5</Terminal>
                  </Edge>
                   <Edge val = "bad">
                      <Terminal>0.5</Terminal>
                   </Edge>
                </Node>
           </Edge>
           <Edge val="s2"><Terminal>0.0</Terminal></Edge>
         </Node>
      </DAG>
  </Parameter>
  </CondProb>
  </InitialStateBelief>
  </pomdpx>
 """
        self.reader_string = PomdpXReader(string=string)
        self.reader_file = PomdpXReader(path=six.StringIO(string))
        expected_dd_parameter = [{
            'Var': 'rover_0',
            'Parent': ['null'],
            'Type': 'DD',
            'Parameter': {'rover_0': {'s0': '0.0',
                                      's1': {'rock_0': {'good': '0.5',
                                                        'bad': '0.5'}},
                                      's2': '0.0'}}}]
        self.maxDiff = None
        self.assertEqual(expected_dd_parameter,
                         self.reader_string.get_initial_beliefs())
        self.assertEqual(expected_dd_parameter,
                         self.reader_file.get_initial_beliefs())

    def test_initial_belief_dd(self):
        string = """
    <pomdpx version="1.0" id="rockSample"
      xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
      xsi:noNamespaceSchemaLocation="pomdpx.xsd">
         <Description>RockSample problem for map size 1 x 3.
           Rock is at 0, Rover’s initial position is at 1.
           Exit is at 2.
         </Description>
         <Discount>0.95</Discount>
        <InitialStateBelief>
        <CondProb>
            <Var>rover_0</Var>
            <Parent>null</Parent>
            <Parameter type="DD">
                <DAG>
                    <Node var="rover_0">
                        <Edge val="s0">
                            <Terminal>0.0</Terminal>
                        </Edge>
                        <Edge val="s1">
                            <SubDAG type="uniform" var="rock_0"/>
                        </Edge>
                        <Edge val="s2">
                            <Terminal>0.0</Terminal>
                        </Edge>
                    </Node>
                </DAG>
            </Parameter>
        </CondProb>
    </InitialStateBelief>
    </pomdpx>
    """
        self.reader_string = PomdpXReader(string=string)
        self.reader_file = PomdpXReader(path=six.StringIO(string))
        expected_belief_dd = [{
            'Var': 'rover_0',
            'Parent': ['null'],
            'Type': 'DD',
            'Parameter': {'rover_0': {'s0': '0.0',
                                      's1': {'type': 'uniform',
                                             'var': 'rock_0'},
                                      's2': '0.0'}}}]
        self.maxDiff = None
        self.assertEqual(self.reader_string.get_initial_beliefs(),
                         expected_belief_dd)
        self.assertEqual(self.reader_file.get_initial_beliefs(),
                         expected_belief_dd)

    def test_reward_function(self):
        string = """
        <pomdpx version="1.0" id="rockSample"
      xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
      xsi:noNamespaceSchemaLocation="pomdpx.xsd">
         <Description>RockSample problem for map size 1 x 3.
           Rock is at 0, Rover’s initial position is at 1.
           Exit is at 2.
         </Description>
         <Discount>0.95</Discount>
        <RewardFunction>
        <Func>
            <Var>reward_rover</Var>
            <Parent>action_rover rover_0 rock_0</Parent>
            <Parameter type="DD">
                <DAG>
                    <Node var="action_rover">
                        <Edge val="amw">
                            <Node var="rover_0">
                                <Edge val="s0">
                                    <Terminal>-100.0</Terminal>
                                </Edge>
                                <Edge val="s1">
                                    <Terminal>0.0</Terminal>
                                </Edge>
                                <Edge val="s2">
                                    <Terminal>0.0</Terminal>
                                </Edge>
                            </Node>
                        </Edge>
                        <Edge val="ame">
                            <Node var="rover_0">
                                <Edge val="s0">
                                    <Terminal>0.0</Terminal>
                                </Edge>
                                <Edge val="s1">
                                    <Terminal>10.0</Terminal>
                                </Edge>
                                <Edge val="s2">
                                    <Terminal>0.0</Terminal>
                                </Edge>
                            </Node>
                        </Edge>
                        <Edge val="ac">
                            <Terminal>0.0</Terminal>
                        </Edge>
                        <Edge val="as">
                            <Node var="rover_0">
                                <Edge val="s0">
                                    <Node var="rock_0">
                                        <Edge val="good">
                                            <Terminal>10</Terminal>
                                        </Edge>
                                        <Edge val="bad">
                                            <Terminal>-10</Terminal>
                                        </Edge>
                                    </Node>
                                </Edge>
                                <Edge val="s1">
                                    <Terminal>-100</Terminal>
                                </Edge>
                                <Edge val="s2">
                                    <Terminal>-100</Terminal>
                                </Edge>
                            </Node>
                        </Edge>
                    </Node>
                </DAG>
            </Parameter>
        </Func>
    </RewardFunction>
    </pomdpx>
        """
        self.reader_string = PomdpXReader(string=string)
        self.reader_file = PomdpXReader(path=six.StringIO(string))
        expected_reward_function_dd =\
            [{'Var': 'reward_rover',
              'Parent': ['action_rover', 'rover_0', 'rock_0'],
              'Type': 'DD',
              'Parameter': {'action_rover': {'amw': {'rover_0': {'s0': '-100.0',
                                                                 's1': '0.0',
                                                                 's2': '0.0'}},
                                             'ame': {'rover_0': {'s0': '0.0',
                                                                 's1': '10.0',
                                                                 's2': '0.0'}},
                                             'ac': '0.0',
                                             'as': {'rover_0': {'s0': {'rock_0': {'good': '10',
                                                                                  'bad': '-10'}},
                                                                's1': '-100',
                                                                's2': '-100'}}}}}]
        self.maxDiff = None
        self.assertEqual(self.reader_string.get_reward_function(),
                         expected_reward_function_dd)
        self.assertEqual(self.reader_file.get_reward_function(),
                         expected_reward_function_dd)

    def test_state_transition_function(self):
        string = """
         <pomdpx version="1.0" id="rockSample"
      xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
      xsi:noNamespaceSchemaLocation="pomdpx.xsd">
         <Description>RockSample problem for map size 1 x 3.
           Rock is at 0, Rover’s initial position is at 1.
           Exit is at 2.
         </Description>
         <Discount>0.95</Discount>
        <StateTransitionFunction>
        <CondProb>
            <Var>rover_1</Var>
            <Parent>action_rover rover_0</Parent>
            <Parameter type="DD">
                <DAG>
                    <Node var="action_rover">
                        <Edge val="amw">
                            <Node var="rover_0">
                                <Edge val="s0">
                                    <SubDAG type="deterministic" var="rover_1" val="s2"/>
                                </Edge>
                                <Edge val="s1">
                                    <SubDAG type="deterministic" var="rover_1" val="s0"/>
                                </Edge>
                                <Edge val="s2">
                                    <SubDAG type="deterministic" var="rover_1" val="s2"/>
                                </Edge>
                            </Node>
                        </Edge>
                        <Edge val="ame">
                            <Node var="rover_0">
                                <Edge val="s0">
                                    <SubDAG type="deterministic" var="rover_1" val="s1"/>
                                </Edge>
                                <Edge val="s1">
                                    <SubDAG type="deterministic" var="rover_1" val="s2"/>
                                </Edge>
                                <Edge val="s2">
                                    <SubDAG type="deterministic" var="rover_1" val="s2"/>
                                </Edge>
                            </Node>
                        </Edge>
                        <Edge val="ac">
                            <SubDAG type="persistent" var="rover_1"/>
                        </Edge>
                        <Edge val="as">
                            <Node var="rover_0">
                                <Edge val="s0">
                                    <SubDAG type="deterministic" var="rover_1" val="s0"/>
                                </Edge>
                                <Edge val="s1">
                                    <SubDAG type="deterministic" var="rover_1" val="s2"/>
                                </Edge>
                                <Edge val="s2">
                                    <SubDAG type="deterministic" var="rover_1" val="s2"/>
                                </Edge>
                            </Node>
                        </Edge>
                    </Node>
                </DAG>
            </Parameter>
        </CondProb>
        <CondProb>
            <Var>rock_1</Var>
            <Parent>action_rover rover_0 rock_0</Parent>
            <Parameter type="DD">
                <DAG>
                    <Node var="action_rover">
                        <Edge val="amw">
                            <SubDAG type="persistent" var="rock_1"/>
                        </Edge>
                        <Edge val="ame">
                            <SubDAG type="persistent" var="rock_1"/>
                        </Edge>
                        <Edge val="ac">
                            <SubDAG type="persistent" var="rock_1"/>
                        </Edge>
                        <Edge val="as">
                            <Node var="rover_0">
                                <Edge val="s0">
                                    <SubDAG type="deterministic" var="rock_1" val="bad"/>
                                </Edge>
                                <Edge val="s1">
                                    <SubDAG type="persistent" var="rock_1"/>
                                </Edge>
                                <Edge val="s2">
                                    <SubDAG type="persistent" var="rock_1"/>
                                </Edge>
                            </Node>
                        </Edge>
                    </Node>
                </DAG>
            </Parameter>
        </CondProb>
    </StateTransitionFunction>
</pomdpx>
        """
        self.reader_string = PomdpXReader(string=string)
        self.reader_file = PomdpXReader(path=six.StringIO(string))
        expected_state_transition_function = \
            [{'Var': 'rover_1',
              'Parent': ['action_rover', 'rover_0'],
              'Type': 'DD',
              'Parameter': {'action_rover': {'amw': {'rover_0': {'s0': {'type': 'deterministic',
                                                                        'var': 'rover_1',
                                                                        'val': 's2'},
                                                                 's1': {'type': 'deterministic',
                                                                        'var': 'rover_1',
                                                                        'val': 's0'},
                                                                 's2': {'type': 'deterministic',
                                                                        'var': 'rover_1',
                                                                        'val': 's2'}}},
                                             'ame': {'rover_0': {'s0': {'type': 'deterministic',
                                                                        'var': 'rover_1',
                                                                        'val': 's1'},
                                                                 's1': {'type': 'deterministic',
                                                                        'var': 'rover_1',
                                                                        'val': 's2'},
                                                                 's2': {'type': 'deterministic',
                                                                        'var': 'rover_1',
                                                                        'val': 's2'},
                                                                 }},
                                             'ac': {'type': 'persistent',
                                                    'var': 'rover_1'},
                                             'as': {'rover_0': {'s0': {'type': 'deterministic',
                                                                       'var': 'rover_1',
                                                                       'val': 's0'},
                                                                's1': {'type': 'deterministic',
                                                                       'var': 'rover_1',
                                                                       'val': 's2'},
                                                                's2': {'type': 'deterministic',
                                                                       'var': 'rover_1',
                                                                       'val': 's2'}}}}}},
             {'Var': 'rock_1',
              'Parent': ['action_rover', 'rover_0', 'rock_0'],
              'Type': 'DD',
              'Parameter': {'action_rover': {'amw': {'type': 'persistent',
                                                     'var': 'rock_1'},
                                             'ame': {'type': 'persistent',
                                                     'var': 'rock_1'},
                                             'ac': {'type': 'persistent',
                                                    'var': 'rock_1'},
                                             'as': {'rover_0': {'s0': {'type': 'deterministic',
                                                                       'var': 'rock_1',
                                                                       'val': 'bad'},
                                                                's1': {'type': 'persistent',
                                                                       'var': 'rock_1'},
                                                                's2': {'type': 'persistent',
                                                                       'var': 'rock_1'}}}}}}]
        self.maxDiff = None
        self.assertEqual(self.reader_string.get_state_transition_function(),
                         expected_state_transition_function)
        self.assertEqual(self.reader_file.get_state_transition_function(),
                         expected_state_transition_function)

    def test_obs_function_dd(self):
        string = """
        <pomdpx version="1.0" id="rockSample"
      xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
      xsi:noNamespaceSchemaLocation="pomdpx.xsd">
         <Description>RockSample problem for map size 1 x 3.
           Rock is at 0, Rover’s initial position is at 1.
           Exit is at 2.
         </Description>
         <Discount>0.95</Discount>
        <ObsFunction>
        <CondProb>
            <Var>obs_sensor</Var>
            <Parent>action_rover rover_1 rock_1</Parent>
            <Parameter type="DD">
                <DAG>
                    <Node var="action_rover">
                        <Edge val="amw">
                            <SubDAG type="deterministic" var="obs_sensor" val="ogood"/>
                        </Edge>
                        <Edge val="ame">
                            <SubDAG type="deterministic" var="obs_sensor" val="ogood"/>
                        </Edge>
                        <Edge val="ac">
                            <Node var="rover_1">
                                <Edge val="s0">
                                    <Node var="rock_1">
                                        <Edge val="good">
                                            <SubDAG type="deterministic" var="obs_sensor" val="ogood"/>
                                        </Edge>
                                        <Edge val="bad">
                                            <SubDAG type="deterministic" var="obs_sensor" val="obad"/>
                                        </Edge>
                                    </Node>
                                </Edge>
                                <Edge val="s1">
                                    <SubDAG type="template" idref="obs_rock"/>
                                </Edge>
                                <Edge val="s2">
                                    <SubDAG type="template" idref="obs_rock"/>
                                </Edge>
                            </Node>
                        </Edge>
                        <Edge val="as">
                            <SubDAG type="deterministic" var="obs_sensor" val="ogood"/>
                        </Edge>
                    </Node>
                </DAG>
                <SubDAGTemplate id="obs_rock">
                    <Node var="rock_1">
                        <Edge val="good">
                            <Node var="obs_sensor">
                                <Edge val="ogood">
                                    <Terminal>0.8</Terminal>
                                </Edge>
                                <Edge val="obad">
                                    <Terminal>0.2</Terminal>
                                </Edge>
                            </Node>
                        </Edge>
                        <Edge val="bad">
                            <Node var="obs_sensor">
                                <Edge val="ogood">
                                    <Terminal>0.2</Terminal>
                                </Edge>
                                <Edge val="obad">
                                    <Terminal>0.8</Terminal>
                                </Edge>
                            </Node>
                        </Edge>
                    </Node>
                </SubDAGTemplate>
            </Parameter>
        </CondProb>
    </ObsFunction>
    </pomdpx>
        """
        self.reader_string = PomdpXReader(string=string)
        self.reader_file = PomdpXReader(path=six.StringIO(string))
        expected_obs_function = \
            [{'Var': 'obs_sensor',
              'Parent': ['action_rover', 'rover_1', 'rock_1'],
              'Type': 'DD',
              'Parameter': {'action_rover': {'amw': {'type': 'deterministic',
                                                     'var': 'obs_sensor',
                                                     'val': 'ogood'},
                                             'ame': {'type': 'deterministic',
                                                     'var': 'obs_sensor',
                                                     'val': 'ogood'},
                                             'ac': {'rover_1': {'s0': {'rock_1': {'good': {'type': 'deterministic',
                                                                                           'var': 'obs_sensor',
                                                                                           'val': 'ogood'},
                                                                                  'bad': {'type': 'deterministic',
                                                                                          'var': 'obs_sensor',
                                                                                          'val': 'obad'}}},
                                                                's1': {'type': 'template',
                                                                       'idref': 'obs_rock'},
                                                                's2': {'type': 'template',
                                                                       'idref': 'obs_rock'}}},
                                             'as': {'type': 'deterministic',
                                                    'var': 'obs_sensor',
                                                    'val': 'ogood'}},
                            'SubDAGTemplate': {'rock_1': {'good': {'obs_sensor': {'ogood': '0.8',
                                                                                  'obad': '0.2'}},
                                                          'bad': {'obs_sensor': {'ogood': '0.2',
                                                                                 'obad': '0.8'}}}},
                            'id': 'obs_rock'}}]
        self.maxDiff = None
        self.assertEqual(self.reader_string.get_obs_function(), expected_obs_function)
        self.assertEqual(self.reader_file.get_obs_function(), expected_obs_function)

    def tearDown(self):
        del self.reader_file
        del self.reader_string


class TestPomdpXWriter(unittest.TestCase):
    def setUp(self):
        self.model_data = {'discription': '',
                           'discount': '0.95',
                           'variables': {
                               'StateVar': [{'vnamePrev': 'rover_0',
                                             'vnameCurr': 'rover_1',
                                             'ValueEnum': ['s0', 's1', 's2'],
                                             'fullyObs': True},
                                            {'vnamePrev': 'rock_0',
                                             'vnameCurr': 'rock_1',
                                             'fullyObs': False,
                                             'ValueEnum': ['good', 'bad']}],
                               'ObsVar': [{'vname': 'obs_sensor',
                                           'ValueEnum': ['ogood', 'obad']}],
                               'RewardVar': [{'vname': 'reward_rover'}],
                               'ActionVar': [{'vname': 'action_rover',
                                              'ValueEnum': ['amw', 'ame',
                                                            'ac', 'as']}]},
                           'initial_state_belief': [{'Var': 'rover_0',
                                                     'Parent': ['null'],
                                                     'Type': 'TBL',
                                                     'Parameter': [{'Instance': ['-'],
                                                                    'ProbTable': ['0.0', '1.0', '0.0']}]},
                                                    {'Var': 'rock_0',
                                                     'Parent': ['null'],
                                                     'Type': 'TBL',
                                                     'Parameter': [{'Instance': ['-'],
                                                                    'ProbTable': ['uniform']}]}],
                           'state_transition_function': [{'Var': 'rover_1',
                                                          'Parent': ['action_rover', 'rover_0'],
                                                          'Type': 'TBL',
                                                          'Parameter': [{'Instance': ['amw', 's0', 's2'],
                                                                         'ProbTable': ['1.0']},
                                                                        {'Instance': ['amw', 's1', 's0'],
                                                                         'ProbTable': ['1.0']},
                                                                        {'Instance': ['ame', 's0', 's1'],
                                                                         'ProbTable': ['1.0']},
                                                                        {'Instance': ['ame', 's1', 's2'],
                                                                         'ProbTable': ['1.0']},
                                                                        {'Instance': ['ac', 's0', 's0'],
                                                                         'ProbTable': ['1.0']},
                                                                        {'Instance': ['ac', 's1', 's1'],
                                                                         'ProbTable': ['1.0']},
                                                                        {'Instance': ['as', 's0', 's0'],
                                                                         'ProbTable': ['1.0']},
                                                                        {'Instance': ['as', 's1', 's2'],
                                                                         'ProbTable': ['1.0']},
                                                                        {'Instance': ['*', 's2', 's2'],
                                                                         'ProbTable': ['1.0']}]},
                                                         {'Var': 'rock_1',
                                                          'Parent': ['action_rover', 'rover_0', 'rock_0'],
                                                          'Type': 'TBL',
                                                          'Parameter': [{'Instance': ['amw', '*', '-', '-'],
                                                                         'ProbTable': ['1.0', '0.0', '0.0', '1.0']},
                                                                        {'Instance': ['ame', '*', '-', '-'],
                                                                         'ProbTable': ['identity']},
                                                                        {'Instance': ['ac', '*', '-', '-'],
                                                                         'ProbTable': ['identity']},
                                                                        {'Instance': ['as', '*', '-', '-'],
                                                                         'ProbTable': ['identity']},
                                                                        {'Instance': ['as', 's0', '*', '-'],
                                                                         'ProbTable': ['0.0', '1.0']},
                                                                        ]}],
                           'obs_function': [{'Var': 'obs_sensor',
                                             'Parent': ['action_rover', 'rover_1', 'rock_1'],
                                             'Type': 'TBL',
                                             'Parameter': [{'Instance': ['amw', '*', '*', '-'],
                                                            'ProbTable': ['1.0', '0.0']},
                                                           {'Instance': ['ame', '*', '*', '-'],
                                                            'ProbTable': ['1.0', '0.0']},
                                                           {'Instance': ['as', '*', '*', '-'],
                                                            'ProbTable': ['1.0', '0.0']},
                                                           {'Instance': ['ac', 's0', '-', '-'],
                                                            'ProbTable': ['1.0', '0.0', '0.0', '1.0']},
                                                           {'Instance': ['ac', 's1', '-', '-'],
                                                            'ProbTable': ['0.8', '0.2', '0.2', '0.8']},
                                                           {'Instance': ['ac', 's2', '*', '-'],
                                                            'ProbTable': ['1.0', '0.0']}]}],
                           'reward_function': [{'Var': 'reward_rover',
                                                'Parent': ['action_rover', 'rover_0', 'rock_0'],
                                                'Type': 'TBL',
                                                'Parameter': [{'Instance': ['ame', 's1', '*'],
                                                               'ValueTable': ['10']},
                                                              {'Instance': ['amw', 's0', '*'],
                                                               'ValueTable': ['-100']},
                                                              {'Instance': ['as', 's1', '*'],
                                                               'ValueTable': ['-100']},
                                                              {'Instance': ['as', 's0', 'good'],
                                                               'ValueTable': ['10']},
                                                              {'Instance': ['as', 's0', 'bad'],
                                                               'ValueTable': ['-10']}]}]}

        self.writer = PomdpXWriter(model_data=self.model_data)

    def test_variables(self):
        expected_variables = etree.XML("""
<Variable>
  <StateVar fullyObs="true" vnameCurr="rover_1" vnamePrev="rover_0">
    <NumValues>3</NumValues>
  </StateVar>
  <StateVar fullyObs="false" vnameCurr="rock_1" vnamePrev="rock_0">
    <ValueEnum>good bad</ValueEnum>
  </StateVar>
  <ObsVar vname="obs_sensor">
    <ValueEnum>ogood obad</ValueEnum>
  </ObsVar>
  <ActionVar vname="action_rover">
    <ValueEnum>amw ame ac as</ValueEnum>
  </ActionVar>
  <RewardVar vname="reward_rover" />
</Variable>""")
        self.maxDiff = None
        self.assertEqual(self.writer.get_variables(),
                         etree.tostring(expected_variables))

    def test_add_initial_belief(self):
        expected_belief_xml = etree.XML("""
<InitialStateBelief>
  <CondProb>
    <Var>rover_0</Var>
    <Parent>null</Parent>
    <Parameter type="TBL">
      <Entry>
        <Instance> - </Instance>
        <ProbTable>0.0 1.0 0.0</ProbTable>
      </Entry>
    </Parameter>
  </CondProb>
  <CondProb>
    <Var>rock_0</Var>
    <Parent>null</Parent>
    <Parameter type="TBL">
      <Entry>
        <Instance> - </Instance>
        <ProbTable>uniform</ProbTable>
      </Entry>
    </Parameter>
  </CondProb>
</InitialStateBelief>""")
        self.maxDiff = None
        self.assertEqual(str(self.writer.add_initial_belief()),
                         str(etree.tostring(expected_belief_xml)))

    def test_add_transition_function(self):
        expected_transition_xml = etree.XML("""
<StateTransitionFunction>
  <CondProb>
    <Var>rover_1</Var>
    <Parent>action_rover rover_0</Parent>
    <Parameter type="TBL">
      <Entry>
        <Instance>amw s0 s2</Instance>
        <ProbTable>1.0</ProbTable>
      </Entry>
      <Entry>
        <Instance>amw s1 s0</Instance>
        <ProbTable>1.0</ProbTable>
      </Entry>
      <Entry>
        <Instance>ame s0 s1</Instance>
        <ProbTable>1.0</ProbTable>
      </Entry>
      <Entry>
        <Instance>ame s1 s2</Instance>
        <ProbTable>1.0</ProbTable>
      </Entry>
      <Entry>
        <Instance>ac s0 s0</Instance>
        <ProbTable>1.0</ProbTable>
      </Entry>
      <Entry>
        <Instance>ac s1 s1</Instance>
        <ProbTable>1.0</ProbTable>
      </Entry>
      <Entry>
        <Instance>as s0 s0</Instance>
        <ProbTable>1.0</ProbTable>
      </Entry>
      <Entry>
        <Instance>as s1 s2</Instance>
        <ProbTable>1.0</ProbTable>
      </Entry>
      <Entry>
        <Instance>* s2 s2</Instance>
        <ProbTable>1.0</ProbTable>
      </Entry>
    </Parameter>
  </CondProb>
  <CondProb>
    <Var>rock_1</Var>
    <Parent>action_rover rover_0 rock_0</Parent>
    <Parameter type="TBL">
      <Entry>
        <Instance>amw * - - </Instance>
        <ProbTable>1.0 0.0 0.0 1.0</ProbTable>
      </Entry>
      <Entry>
        <Instance>ame * - - </Instance>
        <ProbTable>identity</ProbTable>
      </Entry>
      <Entry>
        <Instance>ac * - - </Instance>
        <ProbTable>identity</ProbTable>
      </Entry>
      <Entry>
        <Instance>as * - - </Instance>
        <ProbTable>identity</ProbTable>
      </Entry>
      <Entry>
        <Instance>as s0 * - </Instance>
        <ProbTable>0.0 1.0</ProbTable>
      </Entry>
    </Parameter>
  </CondProb>
</StateTransitionFunction>""")
        self.maxDiff = None
        self.assertEqual(self.writer.add_state_transition_function(),
                         etree.tostring(expected_transition_xml))

    def test_add_obs_function(self):
        expected_obs_xml = etree.XML("""
<ObsFunction>
  <CondProb>
    <Var>obs_sensor</Var>
    <Parent>action_rover rover_1 rock_1</Parent>
    <Parameter type="TBL">
      <Entry>
        <Instance>amw * * - </Instance>
        <ProbTable>1.0 0.0</ProbTable>
      </Entry>
      <Entry>
        <Instance>ame * * - </Instance>
        <ProbTable>1.0 0.0</ProbTable>
      </Entry>
      <Entry>
        <Instance>as * * - </Instance>
        <ProbTable>1.0 0.0</ProbTable>
      </Entry>
      <Entry>
        <Instance>ac s0 - - </Instance>
        <ProbTable>1.0 0.0 0.0 1.0</ProbTable>
      </Entry>
      <Entry>
        <Instance>ac s1 - - </Instance>
        <ProbTable>0.8 0.2 0.2 0.8</ProbTable>
      </Entry>
      <Entry>
        <Instance>ac s2 * - </Instance>
        <ProbTable>1.0 0.0</ProbTable>
      </Entry>
    </Parameter>
  </CondProb>
</ObsFunction>""")
        self.maxDiff = None
        self.assertEqual(self.writer.add_obs_function(),
                         etree.tostring(expected_obs_xml))

    def test_add_reward_function(self):
        expected_reward_xml = etree.XML("""
<RewardFunction>
  <Func>
    <Var>reward_rover</Var>
    <Parent>action_rover rover_0 rock_0</Parent>
    <Parameter type="TBL">
      <Entry>
        <Instance>ame s1 * </Instance>
        <ValueTable>10</ValueTable>
      </Entry>
      <Entry>
        <Instance>amw s0 * </Instance>
        <ValueTable>-100</ValueTable>
      </Entry>
      <Entry>
        <Instance>as s1 * </Instance>
        <ValueTable>-100</ValueTable>
      </Entry>
      <Entry>
        <Instance>as s0 good</Instance>
        <ValueTable>10</ValueTable>
      </Entry>
      <Entry>
        <Instance>as s0 bad</Instance>
        <ValueTable>-10</ValueTable>
      </Entry>
    </Parameter>
  </Func>
</RewardFunction>""")
        self.maxDiff = None
        self.assertEqual(self.writer.add_reward_function(),
                         etree.tostring(expected_reward_xml))

    def test_initial_state_belief_dd(self):
        self.model_data = {'initial_state_belief': [{
            'Var': 'rover_0',
            'Parent': ['null'],
            'Type': 'DD',
            'Parameter': {'rover_0': {'s0': '0.0',
                                      's1': {'type': 'uniform',
                                             'var': 'rock_0'},
                                      's2': '0.0'}}}]}
        self.writer = PomdpXWriter(model_data=self.model_data)
        expected_xml = etree.XML("""
<InitialStateBelief>
  <CondProb>
    <Var>rover_0</Var>
    <Parent>null</Parent>
    <Parameter type="DD">
      <DAG>
        <Node var="rover_0">
          <Edge val="s0">
            <Terminal>0.0</Terminal>
          </Edge>
          <Edge val="s1">
            <SubDAG type="uniform" var="rock_0"/>
          </Edge>
          <Edge val="s2">
            <Terminal>0.0</Terminal>
          </Edge>
        </Node>
      </DAG>
    </Parameter>
  </CondProb>
</InitialStateBelief>""")
        self.maxDiff = None
        self.assertEqual(self.writer.add_initial_belief(),
                         etree.tostring(expected_xml))

    def test_state_transition_function_dd(self):
        self.model_data = {
            'state_transition_function': [{
                'Var': 'rover_1',
                'Parent': ['action_rover', 'rover_0'],
                'Type': 'DD',
                'Parameter': {'action_rover': {
                    'amw': {'rover_0': {'s0': {
                        'type': 'deterministic',
                        'var': 'rover_1',
                        'val': 's2'},
                        's1': {'type': 'deterministic',
                               'var': 'rover_1',
                               'val': 's0'},
                        's2': {'type': 'deterministic',
                               'var': 'rover_1',
                               'val': 's2'}}},
                    'ame': {'rover_0': {'s0': {'type': 'deterministic',
                                               'var': 'rover_1',
                                               'val': 's1'},
                                        's1': {'type': 'deterministic',
                                               'var': 'rover_1',
                                               'val': 's2'},
                                        's2': {'type': 'deterministic',
                                               'var': 'rover_1',
                                               'val': 's2'},
                                        }},
                    'ac': {'type': 'persistent',
                           'var': 'rover_1'},
                    'as': {'rover_0': {'s0': {'type': 'deterministic',
                                              'var': 'rover_1',
                                              'val': 's0'},
                                       's1': {'type': 'deterministic',
                                              'var': 'rover_1',
                                              'val': 's2'},
                                       's2': {'type': 'deterministic',
                                              'var': 'rover_1',
                                              'val': 's2'}}}}}},
                {'Var': 'rock_1',
                 'Parent': ['action_rover', 'rover_0', 'rock_0'],
                 'Type': 'DD',
                 'Parameter': {'action_rover': {
                     'amw': {'type': 'persistent',
                             'var': 'rock_1'},
                     'ame': {'type': 'persistent',
                             'var': 'rock_1'},
                     'ac': {'type': 'persistent',
                            'var': 'rock_1'},
                     'as': {'rover_0': {'s0': {'type': 'deterministic',
                                               'var': 'rock_1',
                                               'val': 'bad'},
                                        's1': {'type': 'persistent',
                                               'var': 'rock_1'},
                                        's2': {'type': 'persistent',
                                               'var': 'rock_1'}}}}}}]}

        self.writer = PomdpXWriter(model_data=self.model_data)
        expected_xml = etree.XML("""
<StateTransitionFunction>
  <CondProb>
    <Var>rover_1</Var>
    <Parent>action_rover rover_0</Parent>
    <Parameter type="DD">
      <DAG>
        <Node var="action_rover">
          <Edge val="ac">
            <SubDAG type="persistent" var="rover_1"/>
          </Edge>
          <Edge val="ame">
            <Node var="rover_0">
              <Edge val="s0">
                <SubDAG type="deterministic" val="s1" var="rover_1"/>
              </Edge>
              <Edge val="s1">
                <SubDAG type="deterministic" val="s2" var="rover_1"/>
              </Edge>
              <Edge val="s2">
                <SubDAG type="deterministic" val="s2" var="rover_1"/>
              </Edge>
            </Node>
          </Edge>
          <Edge val="amw">
            <Node var="rover_0">
              <Edge val="s0">
                <SubDAG type="deterministic" val="s2" var="rover_1"/>
              </Edge>
              <Edge val="s1">
                <SubDAG type="deterministic" val="s0" var="rover_1"/>
              </Edge>
              <Edge val="s2">
                <SubDAG type="deterministic" val="s2" var="rover_1"/>
              </Edge>
            </Node>
          </Edge>
          <Edge val="as">
            <Node var="rover_0">
              <Edge val="s0">
                <SubDAG type="deterministic" val="s0" var="rover_1"/>
              </Edge>
              <Edge val="s1">
                <SubDAG type="deterministic" val="s2" var="rover_1"/>
              </Edge>
              <Edge val="s2">
                <SubDAG type="deterministic" val="s2" var="rover_1"/>
              </Edge>
            </Node>
          </Edge>
        </Node>
      </DAG>
    </Parameter>
  </CondProb>
  <CondProb>
    <Var>rock_1</Var>
    <Parent>action_rover rover_0 rock_0</Parent>
    <Parameter type="DD">
      <DAG>
        <Node var="action_rover">
          <Edge val="ac">
            <SubDAG type="persistent" var="rock_1"/>
          </Edge>
          <Edge val="ame">
            <SubDAG type="persistent" var="rock_1"/>
          </Edge>
          <Edge val="amw">
            <SubDAG type="persistent" var="rock_1"/>
          </Edge>
          <Edge val="as">
            <Node var="rover_0">
              <Edge val="s0">
                <SubDAG type="deterministic" val="bad" var="rock_1"/>
              </Edge>
              <Edge val="s1">
                <SubDAG type="persistent" var="rock_1"/>
              </Edge>
              <Edge val="s2">
                <SubDAG type="persistent" var="rock_1"/>
              </Edge>
            </Node>
          </Edge>
        </Node>
      </DAG>
    </Parameter>
  </CondProb>
</StateTransitionFunction>""")
        self.maxDiff = None
        self.assertEqual(str(self.writer.add_state_transition_function()),
                         str(etree.tostring(expected_xml)))

    def test_obs_function_dd(self):
        self.model_data = {
            'obs_function': [{
                'Var': 'obs_sensor',
                'Parent': ['action_rover', 'rover_1', 'rock_1'],
                'Type': 'DD',
                'Parameter': {'action_rover': {
                    'amw': {'type': 'deterministic',
                            'var': 'obs_sensor',
                            'val': 'ogood'},
                    'ame': {'type': 'deterministic',
                            'var': 'obs_sensor',
                            'val': 'ogood'},
                    'ac': {'rover_1': {'s0': {'rock_1': {'good': {
                        'type': 'deterministic',
                        'var': 'obs_sensor',
                        'val': 'ogood'},
                        'bad': {'type': 'deterministic',
                                'var': 'obs_sensor',
                                'val': 'obad'}}},
                        's1': {'type': 'template',
                               'idref': 'obs_rock'},
                        's2': {'type': 'template',
                               'idref': 'obs_rock'}}},
                    'as': {'type': 'deterministic',
                           'var': 'obs_sensor',
                           'val': 'ogood'}},
                    'SubDAGTemplate': {'rock_1': {'good': {'obs_sensor': {
                        'ogood': '0.8',
                        'obad': '0.2'}},
                        'bad': {'obs_sensor': {
                            'ogood': '0.2',
                            'obad': '0.8'}}}},
                    'id': 'obs_rock'}}]}

        self.writer = PomdpXWriter(model_data=self.model_data)
        expected_xml = etree.XML("""
<ObsFunction>
  <CondProb>
    <Var>obs_sensor</Var>
    <Parent>action_rover rover_1 rock_1</Parent>
    <Parameter type="DD">
      <DAG>
        <Node var="action_rover">
          <Edge val="ac">
            <Node var="rover_1">
              <Edge val="s0">
                <Node var="rock_1">
                  <Edge val="bad">
                    <SubDAG type="deterministic" val="obad" var="obs_sensor" />
                  </Edge>
                  <Edge val="good">
                    <SubDAG type="deterministic" val="ogood" var="obs_sensor"/>
                  </Edge>
                </Node>
              </Edge>
              <Edge val="s1">
                <SubDAG idref="obs_rock" type="template"/>
              </Edge>
              <Edge val="s2">
                <SubDAG idref="obs_rock" type="template"/>
              </Edge>
            </Node>
          </Edge>
          <Edge val="ame">
            <SubDAG type="deterministic" val="ogood" var="obs_sensor"/>
          </Edge>
          <Edge val="amw">
            <SubDAG type="deterministic" val="ogood" var="obs_sensor" />
          </Edge>
          <Edge val="as">
            <SubDAG type="deterministic" val="ogood" var="obs_sensor"/>
          </Edge>
        </Node>
      </DAG>
      <SubDAGTemplate id="obs_rock">
        <Node var="rock_1">
          <Edge val="bad">
            <Node var="obs_sensor">
              <Edge val="obad">
                <Terminal>0.8</Terminal>
              </Edge>
              <Edge val="ogood">
                <Terminal>0.2</Terminal>
              </Edge>
            </Node>
          </Edge>
          <Edge val="good">
            <Node var="obs_sensor">
              <Edge val="obad">
                <Terminal>0.2</Terminal>
              </Edge>
              <Edge val="ogood">
                <Terminal>0.8</Terminal>
              </Edge>
            </Node>
          </Edge>
        </Node>
      </SubDAGTemplate>
    </Parameter>
  </CondProb>
</ObsFunction>""")
        self.maxDiff = None
        self.assertEqual(str(self.writer.add_obs_function()),
                         str(etree.tostring(expected_xml)))

    def test_reward_function_dd(self):
        self.model_data = {
            'reward_function': [{
                'Var': 'reward_rover',
                'Parent': ['action_rover', 'rover_0', 'rock_0'],
                'Type': 'DD',
                'Parameter': {
                    'action_rover': {
                        'amw': {'rover_0': {'s0': '-100.0',
                                            's1': '0.0',
                                            's2': '0.0'}},
                        'ame': {'rover_0': {'s0': '0.0',
                                            's1': '10.0',
                                            's2': '0.0'}},
                        'ac': '0.0',
                        'as': {'rover_0': {'s0': {'rock_0': {'good': '10',
                                                             'bad': '-10'}},
                                           's1': '-100',
                                           's2': '-100'}}}}}]}

        self.writer = PomdpXWriter(model_data=self.model_data)
        expected_xml = etree.XML("""
<RewardFunction>
  <Func>
    <Var>reward_rover</Var>
    <Parent>action_rover rover_0 rock_0</Parent>
    <Parameter type="DD">
      <DAG>
        <Node var="action_rover">
          <Edge val="ac">
            <Terminal>0.0</Terminal>
          </Edge>
          <Edge val="ame">
            <Node var="rover_0">
              <Edge val="s0">
                <Terminal>0.0</Terminal>
              </Edge>
              <Edge val="s1">
                <Terminal>10.0</Terminal>
              </Edge>
              <Edge val="s2">
                <Terminal>0.0</Terminal>
              </Edge>
            </Node>
          </Edge>
          <Edge val="amw">
            <Node var="rover_0">
              <Edge val="s0">
                <Terminal>-100.0</Terminal>
              </Edge>
              <Edge val="s1">
                <Terminal>0.0</Terminal>
              </Edge>
              <Edge val="s2">
                <Terminal>0.0</Terminal>
              </Edge>
            </Node>
          </Edge>
          <Edge val="as">
            <Node var="rover_0">
              <Edge val="s0">
                <Node var="rock_0">
                  <Edge val="bad">
                    <Terminal>-10</Terminal>
                  </Edge>
                  <Edge val="good">
                    <Terminal>10</Terminal>
                  </Edge>
                </Node>
              </Edge>
              <Edge val="s1">
                <Terminal>-100</Terminal>
              </Edge>
              <Edge val="s2">
                <Terminal>-100</Terminal>
              </Edge>
            </Node>
          </Edge>
        </Node>
      </DAG>
    </Parameter>
  </Func>
</RewardFunction>""")
        self.maxDiff = None
        self.assertEqual(self.writer.add_reward_function(),
                         etree.tostring(expected_xml))



from itertools import combinations

import numpy as np

from pyparsing import alphas, Combine, Literal, Optional, nums, Word

from pgmpy.models import BayesianModel, MarkovModel
from pgmpy.factors.discrete import TabularCPD, DiscreteFactor
from pgmpy.extern.six.moves import map, range


class UAIReader(object):
    """
    Class for reading UAI file format from files or strings.
    """
    def __init__(self, path=None, string=None):
        """
        Initialize an instance of UAI reader class

        Parameters
        ----------
        path : file or str
            Path of the file containing UAI information.

        string : str
            String containing UAI information.

        Example
        -------
        >>> reader = UAIReader('TestUai.uai')

        Reference
        ---------
        http://graphmod.ics.uci.edu/uai08/FileFormat
        """
        if path:
            data = open(path)
            self.network = data.read()
        elif string:
            self.network = string
        else:
            raise ValueError("Must specify either path or string.")
        self.grammar = self.get_grammar()
        self.network_type = self.get_network_type()
        self.variables = self.get_variables()
        self.domain = self.get_domain()
        self.edges = self.get_edges()
        self.tables = self.get_tables()

    def get_grammar(self):
        """
        Returns the grammar of the UAI file.
        """
        network_name = Word(alphas).setResultsName('network_name')
        no_variables = Word(nums).setResultsName('no_variables')
        grammar = network_name + no_variables
        self.no_variables = int(grammar.parseString(self.network)['no_variables'])
        domain_variables = (Word(nums)*self.no_variables).setResultsName('domain_variables')
        grammar += domain_variables
        no_functions = Word(nums).setResultsName('no_functions')
        grammar += no_functions
        self.no_functions = int(grammar.parseString(self.network)['no_functions'])
        integer = Word(nums).setParseAction(lambda t: int(t[0]))
        for function in range(0, self.no_functions):
            scope_grammar = Word(nums).setResultsName('fun_scope_' + str(function))
            grammar += scope_grammar
            function_scope = grammar.parseString(self.network)['fun_scope_' + str(function)]
            function_grammar = ((integer)*int(function_scope)).setResultsName('fun_' + str(function))
            grammar += function_grammar

        floatnumber = Combine(Word(nums) + Optional(Literal(".") + Optional(Word(nums))))
        for function in range(0, self.no_functions):
            no_values_grammar = Word(nums).setResultsName('fun_no_values_' + str(function))
            grammar += no_values_grammar
            no_values = grammar.parseString(self.network)['fun_no_values_' + str(function)]
            values_grammar = ((floatnumber)*int(no_values)).setResultsName('fun_values_' + str(function))
            grammar += values_grammar
        return grammar

    def get_network_type(self):
        """
        Returns the type of network defined by the file.

        Returns
        -------
        string : str
            String containing network type.

        Example
        -------
        >>> reader = UAIReader('TestUAI.uai')
        >>> reader.get_network_type()
        'MARKOV'
        """
        network_type = self.grammar.parseString(self.network)
        return network_type['network_name']

    def get_variables(self):
        """
        Returns a list of variables.
        Each variable is represented by an index of list.
        For example if the no of variables are 4 then the list will be
        [var_0, var_1, var_2, var_3]

        Returns
        -------
        list: list of variables

        Example
        -------
        >>> reader = UAIReader('TestUAI.uai')
        >>> reader.get_variables()
        ['var_0', 'var_1', 'var_2']
        """
        variables = []
        for var in range(0, self.no_variables):
            var_name = "var_" + str(var)
            variables.append(var_name)
        return variables

    def get_domain(self):
        """
        Returns the dictionary of variables with keys as variable name
        and values as domain of the variables.

        Returns
        -------
        dict: dictionary containing variables and their domains

        Example
        -------
        >>> reader = UAIReader('TestUAI.uai')
        >>> reader.get_domain()
        {'var_0': '2', 'var_1': '2', 'var_2': '3'}
        """
        domain = {}
        var_domain = self.grammar.parseString(self.network)['domain_variables']
        for var in range(0, len(var_domain)):
            domain["var_" + str(var)] = var_domain[var]
        return domain

    def get_edges(self):
        """
        Returns the edges of the network.

        Returns
        -------
        set: set containing the edges of the network

        Example
        -------
        >>> reader = UAIReader('TestUAI.uai')
        >>> reader.get_edges()
        {('var_0', 'var_1'), ('var_0', 'var_2'), ('var_1', 'var_2')}
        """
        edges = []
        for function in range(0, self.no_functions):
            function_variables = self.grammar.parseString(self.network)['fun_' + str(function)]
            if isinstance(function_variables, int):
                function_variables = [function_variables]
            if self.network_type == 'BAYES':
                child_var = "var_" + str(function_variables[-1])
                function_variables = function_variables[:-1]
                for var in function_variables:
                    edges.append((child_var, "var_" + str(var)))
            elif self.network_type == "MARKOV":
                function_variables = ["var_" + str(var) for var in function_variables]
                edges.extend(list(combinations(function_variables, 2)))
        return set(edges)

    def get_tables(self):
        """
        Returns list of tuple of child variable and CPD in case of Bayesian
        and list of tuple of scope of variables and values in case of Markov.

        Returns
        -------
        list : list of tuples of child variable and values in Bayesian
            list of tuples of scope of variables and values in case of Markov.

        Example
        -------
        >>> reader = UAIReader('TestUAI.uai')
        >>> reader.get_tables()
        [(['var_0', 'var_1'], ['4.000', '2.400', '1.000', '0.000']),
         (['var_0', 'var_1', 'var_2'],
          ['2.2500', '3.2500', '3.7500', '0.0000', '0.0000', '10.0000',
           '1.8750', '4.0000', '3.3330', '2.0000', '2.0000', '3.4000'])]
        """
        tables = []
        for function in range(0, self.no_functions):
            function_variables = self.grammar.parseString(self.network)['fun_' + str(function)]
            if isinstance(function_variables, int):
                function_variables = [function_variables]
            if self.network_type == 'BAYES':
                child_var = "var_" + str(function_variables[-1])
                values = self.grammar.parseString(self.network)['fun_values_' + str(function)]
                tables.append((child_var, list(values)))
            elif self.network_type == "MARKOV":
                function_variables = ["var_" + str(var) for var in function_variables]
                values = self.grammar.parseString(self.network)['fun_values_' + str(function)]
                tables.append((function_variables, list(values)))
        return tables

    def get_model(self):
        """
        Returns an instance of Bayesian Model or Markov Model.
        Varibles are in the pattern var_0, var_1, var_2 where var_0 is
        0th index variable, var_1 is 1st index variable.

        Return
        ------
        model: an instance of Bayesian or Markov Model.

        Examples
        --------
        >>> reader = UAIReader('TestUAI.uai')
        >>> reader.get_model()
        """
        if self.network_type == 'BAYES':
            model = BayesianModel(self.edges)

            tabular_cpds = []
            for cpd in self.tables:
                child_var = cpd[0]
                states = int(self.domain[child_var])
                arr = list(map(float, cpd[1]))
                values = np.array(arr)
                values = values.reshape(states, values.size // states)
                tabular_cpds.append(TabularCPD(child_var, states, values))

            model.add_cpds(*tabular_cpds)
            return model

        elif self.network_type == 'MARKOV':
            model = MarkovModel(self.edges)

            factors = []
            for table in self.tables:
                variables = table[0]
                cardinality = [int(self.domain[var]) for var in variables]
                value = list(map(float, table[1]))
                factor = DiscreteFactor(variables=variables, cardinality=cardinality, values=value)
                factors.append(factor)

            model.add_factors(*factors)
            return model


class UAIWriter(object):
    """
    Class for writing models in UAI.
    """
    def __init__(self, model):
        """
        Initialize an instance of UAI writer class

        Parameters
        ----------
        model: A Bayesian or Markov model
            The model to write
        """
        if isinstance(model, BayesianModel):
            self.network = "BAYES\n"
        elif isinstance(model, MarkovModel):
            self.network = "MARKOV\n"
        else:
            raise TypeError("Model must be an instance of Bayesian or Markov model.")

        self.model = model
        self.no_nodes = self.get_nodes()
        self.domain = self.get_domain()
        self.functions = self.get_functions()
        self.tables = self.get_tables()

    def __str__(self):
        """
        Returns the UAI file as a string.
        """
        self.network += self.no_nodes + "\n"
        domain = sorted(self.domain.items(), key=lambda x: (x[1], x[0]))
        self.network += " ".join([var[1] for var in domain]) + "\n"
        self.network += str(len(self.functions)) + "\n"
        for fun in self.functions:
            self.network += str(len(fun)) + " "
            self.network += " ".join(fun) + "\n"
        self.network += "\n"
        for table in self.tables:
            self.network += str(len(table)) + "\n"
            self.network += " ".join(table) + "\n"
        return self.network[:-1]

    def get_nodes(self):
        """
        Adds variables to the network.

        Example
        -------
        >>> writer = UAIWriter(model)
        >>> writer.get_nodes()
        """
        no_nodes = len(self.model.nodes())
        return str(no_nodes)

    def get_domain(self):
        """
        Adds domain of each variable to the network.

        Example
        -------
        >>> writer = UAIWriter(model)
        >>> writer.get_domain()
        """
        if isinstance(self.model, BayesianModel):
            cpds = self.model.get_cpds()
            cpds.sort(key=lambda x: x.variable)
            domain = {}
            for cpd in cpds:
                domain[cpd.variable] = str(cpd.variable_card)
            return domain
        elif isinstance(self.model, MarkovModel):
            factors = self.model.get_factors()
            domain = {}
            for factor in factors:
                variables = factor.variables
                for var in variables:
                    if var not in domain:
                        domain[var] = str(factor.get_cardinality([var])[var])
            return domain
        else:
            raise TypeError("Model must be an instance of Markov or Bayesian model.")

    def get_functions(self):
        """
        Adds functions to the network.

        Example
        -------
        >>> writer = UAIWriter(model)
        >>> writer.get_functions()
        """
        if isinstance(self.model, BayesianModel):
            cpds = self.model.get_cpds()
            cpds.sort(key=lambda x: x.variable)
            variables = sorted(self.domain.items(), key=lambda x: (x[1], x[0]))
            functions = []
            for cpd in cpds:
                child_var = cpd.variable
                evidence = cpd.variables[:0:-1]
                function = [str(variables.index((var, self.domain[var]))) for var in evidence]
                function.append(str(variables.index((child_var, self.domain[child_var]))))
                functions.append(function)
            return functions
        elif isinstance(self.model, MarkovModel):
            factors = self.model.get_factors()
            functions = []
            variables = sorted(self.domain.items(), key=lambda x: (x[1], x[0]))
            for factor in factors:
                scope = factor.scope()
                function = [str(variables.index((var, self.domain[var]))) for var in scope]
                functions.append(function)
            return functions
        else:
            raise TypeError("Model must be an instance of Markov or Bayesian model.")

    def get_tables(self):
        """
        Adds tables to the network.

        Example
        -------
        >>> writer = UAIWriter(model)
        >>> writer.get_tables()
        """
        if isinstance(self.model, BayesianModel):
            cpds = self.model.get_cpds()
            cpds.sort(key=lambda x: x.variable)
            tables = []
            for cpd in cpds:
                values = list(map(str, cpd.values.ravel()))
                tables.append(values)
            return tables
        elif isinstance(self.model, MarkovModel):
            factors = self.model.get_factors()
            tables = []
            for factor in factors:
                values = list(map(str, factor.values.ravel()))
                tables.append(values)
            return tables
        else:
            raise TypeError("Model must be an instance of Markov or Bayesian model.")

    def write_uai(self, filename):
        """
        Write the xml data into the file.

        Parameters
        ----------
        filename: Name of the file.

        Examples
        -------
        >>> writer = UAIWriter(model)
        >>> writer.write_xmlbif(test_file)
        """
        writer = self.__str__()
        with open(filename, 'w') as fout:
            fout.write(writer)



from collections import namedtuple

import numpy as np
from itertools import combinations, chain
from pgmpy.extern.six.moves import map


State = namedtuple('State', ['var', 'state'])


def cartesian(arrays, out=None):
    """Generate a cartesian product of input arrays.

    Parameters
    ----------
    arrays : list of array-like
        1-D arrays to form the cartesian product of.
    out : ndarray
        Array to place the cartesian product in.

    Returns
    -------
    out : ndarray
        2-D array of shape (M, len(arrays)) containing cartesian products
        formed of input arrays.

    Examples
    --------
    >>> cartesian(([1, 2, 3], [4, 5], [6, 7]))
    array([[1, 4, 6],
           [1, 4, 7],
           [1, 5, 6],
           [1, 5, 7],
           [2, 4, 6],
           [2, 4, 7],
           [2, 5, 6],
           [2, 5, 7],
           [3, 4, 6],
           [3, 4, 7],
           [3, 5, 6],
           [3, 5, 7]])

    """
    arrays = [np.asarray(x) for x in arrays]
    shape = (len(x) for x in arrays)
    dtype = arrays[0].dtype

    ix = np.indices(shape)
    ix = ix.reshape(len(arrays), -1).T

    if out is None:
        out = np.empty_like(ix, dtype=dtype)

    for n, arr in enumerate(arrays):
        out[:, n] = arrays[n][ix[:, n]]

    return out


def sample_discrete(values, weights, size=1):
    """
    Generate a sample of given size, given a probability mass function.

    Parameters
    ----------
    values: numpy.array: Array of all possible values that the random variable
            can take.
    weights: numpy.array or list of numpy.array: Array(s) representing the PMF of the random variable.
    size: int: Size of the sample to be generated.

    Returns
    -------
    numpy.array: of values of the random variable sampled from the given PMF.

    Example
    -------
    >>> import numpy as np
    >>> from pgmpy.utils.mathext import sample_discrete
    >>> values = np.array(['v_0', 'v_1', 'v_2'])
    >>> probabilities = np.array([0.2, 0.5, 0.3])
    >>> sample_discrete(values, probabilities, 10)
    array(['v_1', 'v_1', 'v_0', 'v_1', 'v_2', 'v_0', 'v_1', 'v_1', 'v_1',
      'v_2'], dtype='<U3')
    """
    weights = np.array(weights)
    if weights.ndim == 1:
        return np.random.choice(values, size=size, p=weights)
    else:
        return np.fromiter(map(lambda t: np.random.choice(values, p=t), weights), dtype='int')


def powerset(l):
    """
    Generates all subsets of list `l` (as tuples).

    Example
    -------
    >>> from pgmpy.utils.mathext import powerset
    >>> list(powerset([1,2,3]))
    [(), (1,), (2,), (3,), (1,2), (1,3), (2,3), (1,2,3)]
    """
    return chain.from_iterable(combinations(l, r) for r in range(len(l)+1))



import unittest

import numpy as np
import numpy.testing as np_test
from scipy.special import beta
from scipy.stats import multivariate_normal

from pgmpy.factors.continuous import ContinuousFactor


class TestContinuousFactor(unittest.TestCase):
    def pdf1(self, x, y):
        return np.power(x, 1) * np.power(y, 2) / beta(x, y)

    def pdf2(self, *args):
        return multivariate_normal.pdf(args, [0, 0], [[1, 0], [0, 1]])

    def pdf3(self, x, y, z):
        return z * (np.power(x, 1) * np.power(y, 2)) / beta(x, y)

    def test_class_init(self):
        phi1 = ContinuousFactor(['x', 'y'], self.pdf1)
        self.assertEqual(phi1.scope(), ['x', 'y'])
        self.assertEqual(phi1.pdf, self.pdf1)

        phi2 = ContinuousFactor(['x1', 'x2'], self.pdf2)
        self.assertEqual(phi2.scope(), ['x1', 'x2'])
        self.assertEqual(phi2.pdf, self.pdf2)

        phi3 = ContinuousFactor(['x', 'y', 'z'], self.pdf3)
        self.assertEqual(phi3.scope(), ['x', 'y', 'z'])
        self.assertEqual(phi3.pdf, self.pdf3)

    def test_class_init_typeerror(self):
        self.assertRaises(TypeError, ContinuousFactor, 'x y', self.pdf1)
        self.assertRaises(TypeError, ContinuousFactor, 'x', self.pdf1)

        self.assertRaises(TypeError, ContinuousFactor, 'x1 x2', self.pdf2)
        self.assertRaises(TypeError, ContinuousFactor, 'x1', self.pdf1)

        self.assertRaises(TypeError, ContinuousFactor, 'x y z', self.pdf3)
        self.assertRaises(TypeError, ContinuousFactor, 'x', self.pdf3)

        self.assertRaises(TypeError, ContinuousFactor, set(['x', 'y']), self.pdf1)
        self.assertRaises(TypeError, ContinuousFactor, {'x': 1, 'y': 2}, self.pdf1)

        self.assertRaises(TypeError, ContinuousFactor, set(['x1', 'x2']), self.pdf2)
        self.assertRaises(TypeError, ContinuousFactor, {'x1': 1, 'x2': 2}, self.pdf1)

        self.assertRaises(TypeError, ContinuousFactor, set(['x', 'y', 'z']), self.pdf3)
        self.assertRaises(TypeError, ContinuousFactor, {'x': 1, 'y': 2, 'z': 3}, self.pdf3)

    def test_class_init_valueerror(self):
        self.assertRaises(ValueError, ContinuousFactor, ['x', 'x'], self.pdf1)
        self.assertRaises(ValueError, ContinuousFactor, ['x', 'y', 'y'], self.pdf1)

        self.assertRaises(ValueError, ContinuousFactor, ['x1', 'x1'], self.pdf2)
        self.assertRaises(ValueError, ContinuousFactor, ['x1', 'x2', 'x2'], self.pdf2)

        self.assertRaises(ValueError, ContinuousFactor, ['x', 'x'], self.pdf1)
        self.assertRaises(ValueError, ContinuousFactor, ['x', 'y', 'y', 'z', 'z'], self.pdf1)


class TestContinuousFactorMethods(unittest.TestCase):
    def pdf1(self, x, y):
        return np.power(x, 1) * np.power(y, 2) / beta(x, y)

    def pdf2(self, x1, x2):
        return multivariate_normal.pdf([x1, x2], [0, 0], [[1, 0], [0, 1]])

    def pdf3(self, x, y, z):
        return z * (np.power(x, 1) * np.power(y, 2)) / beta(x, y)

    def pdf4(self, x1, x2, x3):
        return multivariate_normal.pdf([x1, x2, x3], [0, 0, 0],
                                       [[1, 0, 0], [0, 1, 0], [0, 0, 1]])

    def setUp(self):
        self.phi1 = ContinuousFactor(['x', 'y'], self.pdf1)
        self.phi2 = ContinuousFactor(['x1', 'x2'], self.pdf2)
        self.phi3 = ContinuousFactor(['x', 'y', 'z'], self.pdf3)
        self.phi4 = ContinuousFactor(['x1', 'x2', 'x3'], self.pdf4)

    def test_scope(self):
        self.assertEqual(self.phi1.scope(), self.phi1.scope())
        self.assertEqual(self.phi2.scope(), self.phi2.scope())
        self.assertEqual(self.phi3.scope(), self.phi3.scope())

    def test_assignment(self):
        self.assertEqual(self.phi1.assignment(1.212, 2), self.pdf1(1.212, 2))
        self.assertEqual(self.phi2.assignment(1, -2.231), self.pdf2(1, -2.231))
        self.assertEqual(self.phi3.assignment(1.212, 2.213, -3), self.pdf3(1.212, 2.213, -3))

    def test_reduce(self):
        phi1 = self.phi1.copy()
        phi1.reduce([('x', 1)])
        reduced_pdf1 = lambda y: (np.power(1, 1) * np.power(y, 2))/beta(1, y)
        self.assertEqual(phi1.scope(), ['y'])
        for inp in np.random.rand(4):
            self.assertEqual(phi1.pdf(inp), reduced_pdf1(inp))
            self.assertEqual(phi1.pdf(y=inp), reduced_pdf1(inp))

        phi1 = self.phi1.reduce([('x', 1)], inplace=False)
        self.assertEqual(phi1.scope(), ['y'])
        for inp in np.random.rand(4):
            self.assertEqual(phi1.pdf(inp), reduced_pdf1(inp))
            self.assertEqual(phi1.pdf(y=inp), reduced_pdf1(inp))

        phi2 = self.phi2.copy()
        phi2.reduce([('x2', 7.213)])
        reduced_pdf2 = lambda x1: multivariate_normal.pdf([x1, 7.213], [0, 0], [[1, 0], [0, 1]])
        self.assertEqual(phi2.scope(), ['x1'])
        for inp in np.random.rand(4):
            self.assertEqual(phi2.pdf(inp), reduced_pdf2(inp))
            self.assertEqual(phi2.pdf(x1=inp), reduced_pdf2(inp))

        phi2 = self.phi2.reduce([('x2', 7.213)], inplace=False)
        self.assertEqual(phi2.scope(), ['x1'])
        for inp in np.random.rand(4):
            self.assertEqual(phi2.pdf(inp), reduced_pdf2(inp))
            self.assertEqual(phi2.pdf(x1=inp), reduced_pdf2(inp))

        phi3 = self.phi3.copy()
        phi3.reduce([('y', 0.112), ('z', 23)])
        reduced_pdf4 = lambda x: 23*(np.power(x, 1)*np.power(0.112, 2))/beta(x, 0.112)
        self.assertEqual(phi3.scope(), ['x'])
        for inp in np.random.rand(4):
            self.assertEqual(phi3.pdf(inp), reduced_pdf4(inp))
            self.assertEqual(phi3.pdf(x=inp), reduced_pdf4(inp))

        phi3 = self.phi3.copy()
        phi3.reduce([('y', 0.112)])
        reduced_pdf3 = lambda x, z: z*(np.power(x, 1)*np.power(0.112, 2))/beta(x, 0.112)
        self.assertEqual(phi3.scope(), ['x', 'z'])
        for inp in np.random.rand(4, 2):
            self.assertEqual(phi3.pdf(inp[0], inp[1]), reduced_pdf3(inp[0], inp[1]))
            self.assertEqual(phi3.pdf(x=inp[0], z=inp[1]), reduced_pdf3(inp[0], inp[1]))

        phi3 = self.phi3.reduce([('y', 0.112)], inplace=False)
        self.assertEqual(phi3.scope(), ['x', 'z'])
        for inp in np.random.rand(4, 2):
            self.assertEqual(phi3.pdf(inp[0], inp[1]), reduced_pdf3(inp[0], inp[1]))
            self.assertEqual(phi3.pdf(x=inp[0], z=inp[1]), reduced_pdf3(inp[0], inp[1]))
            self.assertEqual(phi3.pdf(inp[0], z=inp[1]), reduced_pdf3(inp[0], inp[1]))

        phi3 = self.phi3.reduce([('y', 0.112), ('z', 23)], inplace=False)
        self.assertEqual(phi3.scope(), ['x'])
        for inp in np.random.rand(4):
            self.assertEqual(phi3.pdf(inp), reduced_pdf4(inp))
            self.assertEqual(phi3.pdf(x=inp), reduced_pdf4(inp))

    def test_reduce_error(self):
        self.assertRaises(TypeError, self.phi1.reduce, 'x1')
        self.assertRaises(TypeError, self.phi1.reduce, set(['x', 'y']))
        self.assertRaises(TypeError, self.phi1.reduce, {'x': 1, 'y': 1})

        self.assertRaises(TypeError, self.phi4.reduce, 'x4')
        self.assertRaises(TypeError, self.phi4.reduce, set(['x1', 'x2', 'x3']))
        self.assertRaises(TypeError, self.phi4.reduce, {'x1': 1, 'x2': 1, 'x3': 1})

        self.assertRaises(ValueError, self.phi1.reduce, [('z', 3)])
        self.assertRaises(ValueError, self.phi1.reduce, [('x', 0), ('y', 1), ('z', 4)])

        self.assertRaises(ValueError, self.phi4.reduce, [('x4', 7)])
        self.assertRaises(ValueError, self.phi4.reduce, [('x1', 1), ('x2', 2), ('x3', 3), ('x4', 4)])

    def test_marginalize(self):
        phi2 = self.phi2.copy()
        phi2.marginalize(['x2'])
        self.assertEqual(phi2.scope(), ['x1'])
        for inp in np.random.rand(4):
            np_test.assert_almost_equal(phi2.pdf(inp),
                                        multivariate_normal.pdf([inp], [0], [[1]]))

        phi2 = self.phi2.marginalize(['x2'], inplace=False)
        self.assertEqual(phi2.scope(), ['x1'])
        for inp in np.random.rand(4):
            np_test.assert_almost_equal(phi2.pdf(inp),
                                        multivariate_normal.pdf([inp], [0], [[1]]))

        phi4 = self.phi4.copy()
        phi4.marginalize(['x2'])

        self.assertEqual(phi4.scope(), ['x1', 'x3'])
        for inp in np.random.rand(4, 2):
            np_test.assert_almost_equal(
                phi4.pdf(inp[0], inp[1]), multivariate_normal.pdf(
                    [inp[0], inp[1]], [0, 0], [[1, 0], [0, 1]]))

        phi4.marginalize(['x3'])
        self.assertEqual(phi4.scope(), ['x1'])
        for inp in np.random.rand(1):
            np_test.assert_almost_equal(phi4.pdf(inp), multivariate_normal.pdf([inp], [0], [[1]]))

        phi4 = self.phi4.marginalize(['x2'], inplace=False)
        self.assertEqual(phi4.scope(), ['x1', 'x3'])
        for inp in np.random.rand(4, 2):
            np_test.assert_almost_equal(phi4.pdf(inp[0], inp[1]),
                                        multivariate_normal.pdf([inp[0], inp[1]], [0, 0], [[1, 0], [0, 1]]))

        phi4 = phi4.marginalize(['x3'], inplace=False)
        self.assertEqual(phi4.scope(), ['x1'])
        for inp in np.random.rand(1):
            np_test.assert_almost_equal(phi4.pdf(inp),
                                        multivariate_normal.pdf([inp], [0], [[1]]))

    def test_marginalize_error(self):
        self.assertRaises(TypeError, self.phi1.marginalize, 'x1')
        self.assertRaises(TypeError, self.phi1.marginalize, set(['x', 'y']))
        self.assertRaises(TypeError, self.phi1.marginalize, {'x': 1, 'y': 1})

        self.assertRaises(TypeError, self.phi4.marginalize, 'x4')
        self.assertRaises(TypeError, self.phi4.marginalize, set(['x1', 'x2', 'x3']))
        self.assertRaises(TypeError, self.phi4.marginalize, {'x1': 1, 'x2': 1, 'x3': 1})

        self.assertRaises(ValueError, self.phi1.marginalize, ['z'])
        self.assertRaises(ValueError, self.phi1.marginalize, ['x', 'y', 'z'])

        self.assertRaises(ValueError, self.phi4.marginalize, ['x4'])
        self.assertRaises(ValueError, self.phi4.marginalize, ['x1', 'x2', 'x3', 'x4'])

    def test_normalize(self):
        def pdf2(x1, x2):
            return 2 * self.pdf2(x1, x2)

        phi2 = ContinuousFactor(['x1', 'x2'], pdf2)
        phi4 = phi2.copy()

        phi4.normalize()
        self.assertEqual(phi4.scope(), phi2.scope())
        for inp in np.random.rand(1, 2):
            np_test.assert_almost_equal(phi4.pdf(inp[0], inp[1]), 2 * self.pdf2(inp[0], inp[1]))

        phi2.normalize()
        self.assertEqual(phi4.scope(), phi2.scope())
        for inp in np.random.rand(1, 2):
            np_test.assert_almost_equal(phi2.pdf(inp[0], inp[1]), phi4.pdf(inp[0], inp[1]))

    def test_operate(self):
        phi1 = self.phi1.copy()
        phi1._operate(self.phi2, 'product')
        self.assertEqual(phi1.scope(), ['x', 'y', 'x1', 'x2'])
        for inp in np.random.rand(4, 4):
            self.assertEqual(phi1.pdf(*inp), self.phi1.pdf(inp[0], inp[1]) * self.phi2.pdf(inp[2], inp[3]))

        phi1 = self.phi1._operate(self.phi2, 'product', inplace=False)
        self.assertEqual(phi1.scope(), ['x', 'y', 'x1', 'x2'])
        for inp in np.random.rand(4, 4):
            self.assertEqual(phi1.pdf(*inp), self.phi1.pdf(inp[0], inp[1]) * self.phi2.pdf(inp[2], inp[3]))

        phi1 = self.phi1 * self.phi2
        self.assertEqual(phi1.scope(), ['x', 'y', 'x1', 'x2'])
        for inp in np.random.rand(4, 4):
            self.assertEqual(phi1.pdf(*inp), self.phi1.pdf(inp[0], inp[1]) * self.phi2.pdf(inp[2], inp[3]))

        phi3 = self.phi3.copy()
        phi3._operate(self.phi1, 'product')
        self.assertEqual(phi3.scope(), ['x', 'y', 'z'])
        for inp in np.random.rand(4, 3):
            self.assertEqual(phi3.pdf(*inp), self.phi3.pdf(*inp) * self.phi1.pdf(inp[0], inp[1]))

        phi3 = self.phi3._operate(self.phi1, 'product', inplace=False)
        self.assertEqual(phi3.scope(), ['x', 'y', 'z'])
        for inp in np.random.rand(4, 3):
            self.assertEqual(phi3.pdf(*inp), self.phi3.pdf(*inp) * self.phi1.pdf(inp[0], inp[1]))

        phi3 = self.phi3 * self.phi1
        self.assertEqual(phi3.scope(), ['x', 'y', 'z'])
        for inp in np.random.rand(4, 3):
            self.assertEqual(phi3.pdf(*inp), self.phi3.pdf(*inp) * self.phi1.pdf(inp[0], inp[1]))

        phi3 = self.phi3.copy()
        phi3._operate(self.phi1, 'divide')
        self.assertEqual(phi3.scope(), ['x', 'y', 'z'])
        for inp in np.random.rand(4, 3):
            self.assertEqual(phi3.pdf(*inp), self.phi3.pdf(*inp) / self.phi1.pdf(inp[0], inp[1]))

        phi3 = self.phi3._operate(self.phi1, 'divide', inplace=False)
        self.assertEqual(phi3.scope(), ['x', 'y', 'z'])
        for inp in np.random.rand(4, 3):
            self.assertEqual(phi3.pdf(*inp), self.phi3.pdf(*inp) / self.phi1.pdf(inp[0], inp[1]))

        phi3 = self.phi3 / self.phi1
        self.assertEqual(phi3.scope(), ['x', 'y', 'z'])
        for inp in np.random.rand(4, 3):
            self.assertEqual(phi3.pdf(*inp), self.phi3.pdf(*inp) / self.phi1.pdf(inp[0], inp[1]))

        phi4 = self.phi4.copy()
        phi4._operate(self.phi2, 'product')
        self.assertEqual(phi4.scope(), ['x1', 'x2', 'x3'])
        for inp in np.random.rand(4, 3):
            self.assertEqual(phi4.pdf(*inp), self.phi4.pdf(*inp) * self.phi2.pdf(inp[0], inp[1]))

        phi4 = self.phi4._operate(self.phi2, 'product', inplace=False)
        self.assertEqual(phi4.scope(), ['x1', 'x2', 'x3'])
        for inp in np.random.rand(4, 3):
            self.assertEqual(phi4.pdf(*inp), self.phi4.pdf(*inp) * self.phi2.pdf(inp[0], inp[1]))

        phi4 = self.phi4 * self.phi2
        self.assertEqual(phi4.scope(), ['x1', 'x2', 'x3'])
        for inp in np.random.rand(4, 3):
            self.assertEqual(phi4.pdf(*inp), self.phi4.pdf(*inp) * self.phi2.pdf(inp[0], inp[1]))

        phi4 = self.phi4.copy()
        phi4._operate(self.phi2, 'divide')
        self.assertEqual(phi4.scope(), ['x1', 'x2', 'x3'])
        for inp in np.random.rand(4, 3):
            self.assertEqual(phi4.pdf(*inp), self.phi4.pdf(*inp) / self.phi2.pdf(inp[0], inp[1]))

        phi4 = self.phi4._operate(self.phi2, 'divide', inplace=False)
        self.assertEqual(phi4.scope(), ['x1', 'x2', 'x3'])
        for inp in np.random.rand(4, 3):
            self.assertEqual(phi4.pdf(*inp), self.phi4.pdf(*inp) / self.phi2.pdf(inp[0], inp[1]))

        phi4 = self.phi4 / self.phi2
        self.assertEqual(phi4.scope(), ['x1', 'x2', 'x3'])
        for inp in np.random.rand(4, 3):
            self.assertEqual(phi4.pdf(*inp), self.phi4.pdf(*inp) / self.phi2.pdf(inp[0], inp[1]))

    def test_operate_error(self):
        self.assertRaises(TypeError, self.phi1._operate, 1, 'product')
        self.assertRaises(TypeError, self.phi1._operate, 1, 'divide')
        self.assertRaises(TypeError, self.phi1._operate, '1', 'product')
        self.assertRaises(TypeError, self.phi1._operate, '1', 'divide')
        self.assertRaises(TypeError, self.phi1._operate, self.phi2.pdf, 'product')
        self.assertRaises(TypeError, self.phi1._operate, self.phi2.pdf, 'divide')
        self.assertRaises(TypeError, self.phi1._operate, [1], 'product')
        self.assertRaises(TypeError, self.phi1._operate, [1], 'divide')

        self.assertRaises(TypeError, self.phi4._operate, 1, 'product')
        self.assertRaises(TypeError, self.phi4._operate, 1, 'divide')
        self.assertRaises(TypeError, self.phi4._operate, '1', 'product')
        self.assertRaises(TypeError, self.phi4._operate, '1', 'divide')
        self.assertRaises(TypeError, self.phi4._operate, self.phi2.pdf, 'product')
        self.assertRaises(TypeError, self.phi4._operate, self.phi2.pdf, 'divide')
        self.assertRaises(TypeError, self.phi4._operate, [1], 'product')
        self.assertRaises(TypeError, self.phi4._operate, [1], 'divide')

        self.assertRaises(TypeError, self.phi1._operate, 1, 'product', False)
        self.assertRaises(TypeError, self.phi1._operate, 1, 'divide', False)
        self.assertRaises(TypeError, self.phi1._operate, '1', 'product', False)
        self.assertRaises(TypeError, self.phi1._operate, '1', 'divide', False)
        self.assertRaises(TypeError, self.phi1._operate, self.phi2.pdf, 'product', False)
        self.assertRaises(TypeError, self.phi1._operate, self.phi2.pdf, 'divide', False)
        self.assertRaises(TypeError, self.phi1._operate, [1], 'product', False)
        self.assertRaises(TypeError, self.phi1._operate, [1], 'divide', False)

        self.assertRaises(TypeError, self.phi4._operate, 1, 'product', False)
        self.assertRaises(TypeError, self.phi4._operate, 1, 'divide', False)
        self.assertRaises(TypeError, self.phi4._operate, '1', 'product', False)
        self.assertRaises(TypeError, self.phi4._operate, '1', 'divide', False)
        self.assertRaises(TypeError, self.phi4._operate, self.phi2.pdf, 'product', False)
        self.assertRaises(TypeError, self.phi4._operate, self.phi2.pdf, 'divide', False)
        self.assertRaises(TypeError, self.phi4._operate, [1], 'product', False)
        self.assertRaises(TypeError, self.phi4._operate, [1], 'divide', False)

        self.assertRaises(ValueError, self.phi1.__truediv__, self.phi2)
        self.assertRaises(ValueError, self.phi1.__truediv__, self.phi3)
        self.assertRaises(ValueError, self.phi1.__truediv__, self.phi4)
        self.assertRaises(ValueError, self.phi2.__truediv__, self.phi3)
        self.assertRaises(ValueError, self.phi2.__truediv__, self.phi4)

    def test_copy(self):
        copy1 = self.phi1.copy()
        copy2 = self.phi3.copy()

        copy4 = copy1.copy()
        copy5 = copy2.copy()

        self.assertEqual(copy1.scope(), copy4.scope())
        self.assertEqual(copy1.pdf, copy4.pdf)
        self.assertEqual(copy2.scope(), copy5.scope())
        self.assertEqual(copy2.pdf, copy5.pdf)

#TODO: Fix these
#        copy1.variables = ['A', 'B']
#        self.assertEqual(copy4.scope(), self.phi1.scope())

#        def pdf(a, b):
#            return (a + b) / (a * a + b * b)
#        copy1._pdf = pdf
#        copy1_pdf = pdf
#        self.assertEqual(copy4.pdf, self.phi1.pdf)
#        copy4.variables = ['X', 'Y']
#        self.assertEqual(copy1.scope(), ['x', 'y'])
#        copy4._pdf = lambda a, b: a + b
#        for inp in np.random.rand(4, 2):
#            self.assertEqual(copy1.pdf(inp[0], inp[1]), copy1_pdf(inp[0], inp[1]))

#        copy2.reduce([('x', 7.7)])

#        def reduced_pdf(y, z):
#            return z*(np.power(7.7, 1) * np.power(y, 2)) / beta(7.7, y)
#        self.assertEqual(copy5.scope(), self.phi3.scope())
#        self.assertEqual(copy5.pdf, self.phi3.pdf)
#        copy5.reduce([('x', 11), ('z', 13)])
#        self.assertEqual(copy2.scope(), ['y', 'z'])
#        for inp in np.random.rand(4, 2):
#            self.assertEqual(copy2.pdf(inp[0], inp[1]), reduced_pdf(inp[0], inp[1]))

    def tearDown(self):
        del self.phi1
        del self.phi2
        del self.phi3



import unittest

import numpy.testing as np_test

from pgmpy.factors.continuous import LinearGaussianCPD


class TestLGCPD(unittest.TestCase):
    @unittest.skip("TODO")
    def test_class_init(self):
        cpd1 = LinearGaussianCPD('x', [0.23], 0.56)
        self.assertEqual(cpd1.variable, 'x')
        self.assertEqual(cpd1.beta_0, 0.23)
        self.assertEqual(cpd1.variance, 0.56)

        cpd2 = LinearGaussianCPD('y', [0.67, 1, 4.56, 8], 2,
                                 ['x1', 'x2', 'x3'])
        self.assertEqual(cpd2.variable, 'y')
        self.assertEqual(cpd2.beta_0, 0.67)
        self.assertEqual(cpd2.variance, 2)
        self.assertEqual(cpd2.evidence, ['x1', 'x2', 'x3'])
        np_test.assert_array_equal(cpd2.beta_vector, [1, 4.56, 8])

        self.assertRaises(ValueError, LinearGaussianCPD, 'x', [1, 1, 2], 2,
                          ['a', 'b', 'c'])
        self.assertRaises(ValueError, LinearGaussianCPD, 'x', [1, 1, 2, 3], 2,
                          ['a', 'b'])
    @unittest.skip("TODO")
    def test_pdf(self):
        cpd1 = LinearGaussianCPD('x', [0.23], 0.56)
        cpd2 = LinearGaussianCPD('y', [0.67, 1, 4.56, 8], 2, ['x1', 'x2', 'x3'])
        np_test.assert_almost_equal(cpd1.assignment(1), 0.3139868)
        np_test.assert_almost_equal(cpd2.assignment(1, 1.2, 2.3, 3.4), 1.076e-162)

    @unittest.skip("TODO")
    def test_copy(self):
        cpd = LinearGaussianCPD('y', [0.67, 1, 4.56, 8], 2, ['x1', 'x2', 'x3'])
        copy = cpd.copy()

        self.assertEqual(cpd.variable, copy.variable)
        self.assertEqual(cpd.beta_0, copy.beta_0)
        self.assertEqual(cpd.variance, copy.variance)
        np_test.assert_array_equal(cpd.beta_vector, copy.beta_vector)
        self.assertEqual(cpd.evidence, copy.evidence)

        cpd.variable = 'z'
        self.assertEqual(copy.variable, 'y')
        cpd.variance = 0
        self.assertEqual(copy.variance, 2)
        cpd.beta_0 = 1
        self.assertEqual(copy.beta_0, 0.67)
        cpd.evidence = ['p', 'q', 'r']
        self.assertEqual(copy.evidence, ['x1', 'x2', 'x3'])
        cpd.beta_vector = [2, 2, 2]
        np_test.assert_array_equal(copy.beta_vector, [1, 4.56, 8])

        copy = cpd.copy()

        copy.variable = 'k'
        self.assertEqual(cpd.variable, 'z')
        copy.variance = 0.3
        self.assertEqual(cpd.variance, 0)
        copy.beta_0 = 1.5
        self.assertEqual(cpd.beta_0, 1)
        copy.evidence = ['a', 'b', 'c']
        self.assertEqual(cpd.evidence, ['p', 'q', 'r'])
        copy.beta_vector = [2.2, 2.2, 2.2]
        np_test.assert_array_equal(cpd.beta_vector, [2, 2, 2])

    @unittest.skip("TODO")
    def test_str(self):
        cpd1 = LinearGaussianCPD('x', [0.23], 0.56)
        cpd2 = LinearGaussianCPD('y', [0.67, 1, 4.56, 8], 2, ['x1', 'x2', 'x3'])
        self.assertEqual(cpd1.__str__(), "P(x) = N(0.23; 0.56)")
        self.assertEqual(cpd2.__str__(), "P(y | x1, x2, x3) = N(1.0*x1 + "
                                         "4.56*x2 + 8.0*x3 + 0.67; 2)")



from .ProbModelXML import (ProbModelXMLReader, ProbModelXMLWriter, generate_probmodelxml,
                           get_probmodel_data, parse_probmodelxml, write_probmodelxml, read_probmodelxml)
from .XMLBIF import XMLBIFReader, XMLBIFWriter
from .PomdpX import PomdpXReader, PomdpXWriter
from .XMLBeliefNetwork import XBNReader, XBNWriter
from .UAI import UAIReader, UAIWriter
from .BIF import BIFReader, BIFWriter

__all__ = ['ProbModelXMLReader',
           'ProbModelXMLWriter',
           'generate_probmodelxml',
           'get_probmodel_data',
           'parse_probmodelxml',
           'write_probmodelxml',
           'read_probmodelxml',
           'XMLBIFReader',
           'XMLBIFWriter',
           'XBNReader',
           'XBNWriter',
           'PomdpXReader',
           'PomdpXWriter',
           'UAIReader',
           'UAIWriter',
           'BIFReader',
           'BIFWriter']



#!/usr/bin/env python3

import networkx as nx

from pgmpy.models import ClusterGraph


class JunctionTree(ClusterGraph):
    """
    Class for representing Junction Tree.

    Junction tree is undirected graph where each node represents a clique
    (list, tuple or set of nodes) and edges represent sepset between two cliques.
    Each sepset in G separates the variables strictly on one side of edge to
    other.

    Parameters
    ----------
    data: input graph
        Data to initialize graph. If data=None (default) an empty graph is
        created. The data is an edge list.

    Examples
    --------
    Create an empty JunctionTree with no nodes and no edges

    >>> from pgmpy.models import JunctionTree
    >>> G = JunctionTree()

    G can be grown by adding clique nodes.

    **Nodes:**

    Add a tuple (or list or set) of nodes as single clique node.

    >>> G.add_node(('a', 'b', 'c'))
    >>> G.add_nodes_from([('a', 'b'), ('a', 'b', 'c')])

    **Edges:**

    G can also be grown by adding edges.

    >>> G.add_edge(('a', 'b', 'c'), ('a', 'b'))

    or a list of edges

    >>> G.add_edges_from([(('a', 'b', 'c'), ('a', 'b')),
    ...                   (('a', 'b', 'c'), ('a', 'c'))])
    """

    def __init__(self, ebunch=None):
        super(JunctionTree, self).__init__()
        if ebunch:
            self.add_edges_from(ebunch)

    def add_edge(self, u, v, **kwargs):
        """
        Add an edge between two clique nodes.

        Parameters
        ----------
        u, v: nodes
            Nodes can be any list or set or tuple of nodes forming a clique.

        Examples
        --------
        >>> from pgmpy.models import JunctionTree
        >>> G = JunctionTree()
        >>> G.add_nodes_from([('a', 'b', 'c'), ('a', 'b'), ('a', 'c')])
        >>> G.add_edges_from([(('a', 'b', 'c'), ('a', 'b')),
        ...                   (('a', 'b', 'c'), ('a', 'c'))])
        """
        if u in self.nodes() and v in self.nodes() and nx.has_path(self, u, v):
            raise ValueError('Addition of edge between {u} and {v} forms a cycle breaking the '
                             'properties of Junction Tree'.format(u=str(u), v=str(v)))

        super(JunctionTree, self).add_edge(u, v, **kwargs)

    def check_model(self):
        """
        Check the model for various errors. This method checks for the following
        errors. In the same time also updates the cardinalities of all the random
        variables.

        * Checks if clique potentials are defined for all the cliques or not.
        * Check for running intersection property is not done explicitly over
        here as it done in the add_edges method.

        Returns
        -------
        check: boolean
            True if all the checks are passed
        """
        if not nx.is_connected(self):
            raise ValueError('The Junction Tree defined is not fully connected.')

        return super(JunctionTree, self).check_model()

    def copy(self):
        """
        Returns a copy of JunctionTree.

        Returns
        -------
        JunctionTree : copy of JunctionTree

        Examples
        -------
        >>> import numpy as np
        >>> from pgmpy.factors.discrete import DiscreteFactor
        >>> from pgmpy.models import JunctionTree
        >>> G = JunctionTree()
        >>> G.add_edges_from([(('a', 'b', 'c'), ('a', 'b')), (('a', 'b', 'c'), ('a', 'c'))])
        >>> phi1 = DiscreteFactor(['a', 'b'], [1, 2], np.random.rand(2))
        >>> phi2 = DiscreteFactor(['a', 'c'], [1, 2], np.random.rand(2))
        >>> G.add_factors(phi1,phi2)
        >>> modelCopy = G.copy()
        >>> modelCopy.edges()
        [(('a', 'b'), ('a', 'b', 'c')), (('a', 'c'), ('a', 'b', 'c'))]
        >>> G.factors
        [<DiscreteFactor representing phi(a:1, b:2) at 0xb720ee4c>, <DiscreteFactor representing phi(a:1, c:2) at 0xb4e1e06c>]
        >>> modelCopy.factors
        [<DiscreteFactor representing phi(a:1, b:2) at 0xb4bd11ec>, <DiscreteFactor representing phi(a:1, c:2) at 0xb4bd138c>]

        """
        copy = JunctionTree(self.edges())
        copy.add_nodes_from(self.nodes())
        if self.factors:
            factors_copy = [factor.copy() for factor in self.factors]
            copy.add_factors(*factors_copy)
        return copy



from .BayesianModel import BayesianModel
from .ClusterGraph import ClusterGraph
from .DynamicBayesianNetwork import DynamicBayesianNetwork
from .FactorGraph import FactorGraph
from .JunctionTree import JunctionTree
from .MarkovChain import MarkovChain
from .MarkovModel import MarkovModel
from .NaiveBayes import NaiveBayes
from .NoisyOrModel import NoisyOrModel
from .LinearGaussianBayesianNetwork import LinearGaussianBayesianNetwork

__all__ = ['BayesianModel',
           'NoisyOrModel',
           'MarkovModel',
           'FactorGraph',
           'JunctionTree',
           'ClusterGraph',
           'DynamicBayesianNetwork',
           'MarkovChain',
           'NaiveBayes',
           'LinearGaussianBayesianNetwork']



import unittest

import pandas as pd
import numpy as np

from pgmpy.estimators import ConstraintBasedEstimator
from pgmpy.independencies import Independencies
from pgmpy.models import BayesianModel
from pgmpy.base import DirectedGraph, UndirectedGraph


class TestConstraintBasedEstimator(unittest.TestCase):
    def test_build_skeleton(self):
        ind = Independencies(['B', 'C'], ['A', ['B', 'C'], 'D'])
        ind = ind.closure()
        skel1, sep_sets1 = ConstraintBasedEstimator.build_skeleton("ABCD", ind)
        self.assertTrue(self._edge_list_equal(skel1.edges(), [('A', 'D'), ('B', 'D'), ('C', 'D')]))

        sep_sets_ref1 = {frozenset({'A', 'C'}): (), frozenset({'A', 'B'}): (), frozenset({'C', 'B'}): ()}
        self.assertEqual(sep_sets1, sep_sets_ref1)

        model = BayesianModel([('A', 'C'), ('B', 'C'), ('B', 'D'), ('C', 'E')])
        skel2, sep_sets2 = ConstraintBasedEstimator.build_skeleton(model.nodes(), model.get_independencies())
        self.assertTrue(self._edge_list_equal(skel2, [('D', 'B'), ('A', 'C'), ('B', 'C'), ('C', 'E')]))

        sep_sets_ref2 = {frozenset({'D', 'C'}): ('B',),
                         frozenset({'E', 'B'}): ('C',),
                         frozenset({'A', 'D'}): (),
                         frozenset({'E', 'D'}): ('C',),
                         frozenset({'E', 'A'}): ('C',),
                         frozenset({'A', 'B'}): ()}
        # witnesses/seperators might change on each run, so we cannot compare directly
        self.assertEqual(sep_sets2.keys(), sep_sets_ref2.keys())
        self.assertEqual([len(v) for v in sorted(sep_sets2.values())],
                         [len(v) for v in sorted(sep_sets_ref2.values())])

    def test_skeleton_to_pdag(self):
        data = pd.DataFrame(np.random.randint(0, 3, size=(1000, 3)), columns=list('ABD'))
        data['C'] = data['A'] - data['B']
        data['D'] += data['A']
        c = ConstraintBasedEstimator(data)
        pdag = c.skeleton_to_pdag(*c.estimate_skeleton())
        self.assertSetEqual(set(pdag.edges()),
                            set([('B', 'C'), ('A', 'D'), ('A', 'C'), ('D', 'A')]))

        skel = UndirectedGraph([('A', 'B'), ('A', 'C')])
        sep_sets1 = {frozenset({'B', 'C'}): ()}
        self.assertSetEqual(set(c.skeleton_to_pdag(skel, sep_sets1).edges()),
                            set([('B', 'A'), ('C', 'A')]))

        sep_sets2 = {frozenset({'B', 'C'}): ('A',)}
        pdag2 = c.skeleton_to_pdag(skel, sep_sets2)
        self.assertSetEqual(set(c.skeleton_to_pdag(skel, sep_sets2).edges()),
                            set([('A', 'B'), ('B', 'A'), ('A', 'C'), ('C', 'A')]))

    def test_pdag_to_dag(self):
        pdag1 = DirectedGraph([('A', 'B'), ('C', 'B'), ('C', 'D'), ('D', 'C'), ('D', 'A'), ('A', 'D')])
        dag1 = ConstraintBasedEstimator.pdag_to_dag(pdag1)
        self.assertTrue(('A', 'B') in dag1.edges() and
                        ('C', 'B') in dag1.edges() and
                        len(dag1.edges()) == 4)

        pdag2 = DirectedGraph([('B', 'C'), ('D', 'A'), ('A', 'D'), ('A', 'C')])
        dag2 = ConstraintBasedEstimator.pdag_to_dag(pdag2)
        self.assertTrue(set(dag2.edges()) == set([('B', 'C'), ('A', 'D'), ('A', 'C')]) or
                        set(dag2.edges()) == set([('B', 'C'), ('D', 'A'), ('A', 'C')]))

        pdag3 = DirectedGraph([('B', 'C'), ('D', 'C'), ('C', 'D'), ('A', 'C')])
        dag3 = ConstraintBasedEstimator.pdag_to_dag(pdag3)
        self.assertSetEqual(set([('B', 'C'), ('C', 'D'), ('A', 'C')]),
                            set(dag3.edges()))

    def test_estimate_from_independencies(self):
        ind = Independencies(['B', 'C'], ['A', ['B', 'C'], 'D'])
        ind = ind.closure()
        model = ConstraintBasedEstimator.estimate_from_independencies("ABCD", ind)

        self.assertSetEqual(set(model.edges()),
                            set([('B', 'D'), ('A', 'D'), ('C', 'D')]))

        model1 = BayesianModel([('A', 'C'), ('B', 'C'), ('B', 'D'), ('C', 'E')])
        model2 = ConstraintBasedEstimator.estimate_from_independencies(
                            model1.nodes(),
                            model1.get_independencies())

        self.assertTrue(set(model2.edges()) == set(model1.edges()) or
                        set(model2.edges()) == set([('B', 'C'), ('A', 'C'), ('C', 'E'), ('D', 'B')]))

    def test_estimate_skeleton(self):
        data = pd.DataFrame(np.random.randint(0, 2, size=(1000, 5)), columns=list('ABCDE'))
        data['F'] = data['A'] + data['B'] + data['C']
        est = ConstraintBasedEstimator(data)
        skel, sep_sets = est.estimate_skeleton()
        self.assertTrue(self._edge_list_equal(skel.edges(),
                                              [('A', 'F'), ('B', 'F'), ('C', 'F')]))

        sep_sets_ref = {frozenset({'D', 'F'}): (),
                        frozenset({'D', 'B'}): (),
                        frozenset({'A', 'C'}): (),
                        frozenset({'D', 'E'}): (),
                        frozenset({'E', 'F'}): (),
                        frozenset({'E', 'C'}): (),
                        frozenset({'E', 'B'}): (),
                        frozenset({'D', 'C'}): (),
                        frozenset({'A', 'B'}): (),
                        frozenset({'A', 'E'}): (),
                        frozenset({'B', 'C'}): (),
                        frozenset({'A', 'D'}): ()}
        self.assertEqual(set(sep_sets.keys()), set(sep_sets_ref.keys()))

    def test_estimate_skeleton2(self):
        data = pd.DataFrame(np.random.randint(0, 2, size=(1000, 3)), columns=list('XYZ'))
        data['X'] += data['Z']
        data['Y'] += data['Z']
        est = ConstraintBasedEstimator(data)
        skel, sep_sets = est.estimate_skeleton()

        self.assertTrue(self._edge_list_equal(skel.edges(), [('X', 'Z'), ('Y', 'Z')]))
        self.assertEqual(sep_sets, {frozenset(('X', 'Y')): ('Z',)})

    def test_estimate(self):
        data = pd.DataFrame(np.random.randint(0, 3, size=(1000, 3)), columns=list('XYZ'))
        data['sum'] = data.sum(axis=1)
        model = ConstraintBasedEstimator(data).estimate()
        self.assertSetEqual(set(model.edges()),
                            set([('Z', 'sum'), ('X', 'sum'), ('Y', 'sum')]))

    @staticmethod
    def _edge_list_equal(edges1, edges2):
        "Checks if two lists of undirected edges are equal."
        return ((((X, Y) in edges2 or (Y, X) in edges2) for X, Y in edges1) and
                (((X, Y) in edges1 or (Y, X) in edges1) for X, Y in edges2))



import unittest
import numpy as np
import numpy.testing as np_test
from pgmpy.models import NoisyOrModel


class TestNoisyOrModelInit(unittest.TestCase):

    def test_init(self):
        model = NoisyOrModel(['x1', 'x2', 'x3'], [2, 3, 2], [[0.6, 0.4],
                                                             [0.2, 0.4, 0.7],
                                                             [0.1, 0.4]])
        np_test.assert_array_equal(model.variables, np.array(['x1', 'x2', 'x3']))
        np_test.assert_array_equal(model.cardinality, np.array([2, 3, 2]))
        self.assertListEqual(model.inhibitor_probability, [[0.6, 0.4],
                                                           [0.2, 0.4, 0.7],
                                                           [0.1, 0.4]])

    def test_exceptions(self):
        self.assertRaises(ValueError, NoisyOrModel, np.array(['x1', 'x2', 'x3']), [2, 2, 2], [[0.1, 0.2],
                                                                                              [1.0, 0.3],
                                                                                              [1.2, 0.1]])
        self.assertRaises(ValueError, NoisyOrModel, np.array(['x1', 'x2', 'x3']), [2, 4], [[0.1, 0.2],
                                                                                           [0.1, 0.4, 0.2, 0.6]])
        self.assertRaises(ValueError, NoisyOrModel, np.array(['x1', 'x2', 'x3']), [2, 3, 2, 3], [[0.1, 0.2],
                                                                                                 [0.6, 0.3, 0.5],
                                                                                                 [0.3, 0.2],
                                                                                                 [0.1, 0.4, 0.3]])
        self.assertRaises(ValueError, NoisyOrModel, np.array(['x1', 'x2', 'x3']), [2, 3, 2], [[0.1, 0.2, 0.4],
                                                                                              [0.4, 0.1, 0.5],
                                                                                              [0.6, 0.1, 0.7]])
        self.assertRaises(ValueError, NoisyOrModel, np.array(['x1', 'x2', 'x3']), [2, 2, 2], [[0.1, 0.1],
                                                                                              [0.1, 0.1]])


class TestNoisyOrModelMethods(unittest.TestCase):

    def setUp(self):
        self.model = NoisyOrModel(['x1', 'x2', 'x3'], [2, 3, 2], [[0.6, 0.4],
                                                                  [0.2, 0.4, 0.7],
                                                                  [0.1, 0.4]])

    def test_add_variables(self):
        self.model.add_variables(['x4'], [3], [0.1, 0.2, 0.4])
        np_test.assert_array_equal(self.model.variables, np.array(['x1', 'x2', 'x3', 'x4']))
        np_test.assert_array_equal(self.model.cardinality, np.array([2, 3, 2, 3]))
        self.assertListEqual(self.model.inhibitor_probability, [[0.6, 0.4],
                                                                [0.2, 0.4, 0.7],
                                                                [0.1, 0.4],
                                                                [0.1, 0.2, 0.4]])

        self.model.add_variables(['x5', 'x6'], [3, 2], [[0.1, 0.2, 0.4], [0.5, 0.5]])
        np_test.assert_array_equal(self.model.variables, np.array(['x1', 'x2', 'x3', 'x4', 'x5', 'x6']))
        np_test.assert_array_equal(self.model.cardinality, np.array([2, 3, 2, 3, 3, 2]))
        self.assertListEqual(self.model.inhibitor_probability, [[0.6, 0.4],
                                                                [0.2, 0.4, 0.7],
                                                                [0.1, 0.4],
                                                                [0.1, 0.2, 0.4],
                                                                [0.1, 0.2, 0.4],
                                                                [0.5, 0.5]])

    def test_del_variables(self):
        self.model.del_variables(['x3'])
        np_test.assert_array_equal(self.model.variables, np.array(['x1', 'x2']))
        np_test.assert_array_equal(self.model.cardinality, np.array([2, 3]))
        self.assertListEqual(self.model.inhibitor_probability, [[0.6, 0.4],
                                                                [0.2, 0.4, 0.7]])

    def test_del_multiple_variables(self):
        self.model.del_variables(['x1', 'x2'])
        np_test.assert_array_equal(self.model.variables, np.array(['x3']))
        np_test.assert_array_equal(self.model.cardinality, np.array([2]))
        self.assertListEqual(self.model.inhibitor_probability, [[0.1, 0.4]])



import unittest

from pgmpy.independencies import Independencies, IndependenceAssertion


class TestIndependenceAssertion(unittest.TestCase):
    def setUp(self):
        self.assertion = IndependenceAssertion()

    def test_return_list_if_str(self):
        self.assertListEqual(self.assertion._return_list_if_str('U'), ['U'])
        self.assertListEqual(self.assertion._return_list_if_str(['U', 'V']), ['U', 'V'])

    def test_get_assertion(self):
        self.assertTupleEqual(IndependenceAssertion('U', 'V', 'Z').get_assertion(), ({'U'}, {'V'}, {'Z'}))
        self.assertTupleEqual(IndependenceAssertion('U', 'V').get_assertion(), ({'U'}, {'V'}, set()))

    def test_init(self):
        self.assertion1 = IndependenceAssertion('U', 'V', 'Z')
        self.assertSetEqual(self.assertion1.event1, {'U'})
        self.assertSetEqual(self.assertion1.event2, {'V'})
        self.assertSetEqual(self.assertion1.event3, {'Z'})
        self.assertion1 = IndependenceAssertion(['U', 'V'], ['Y', 'Z'], ['A', 'B'])
        self.assertSetEqual(self.assertion1.event1, {'U', 'V'})
        self.assertSetEqual(self.assertion1.event2, {'Y', 'Z'})
        self.assertSetEqual(self.assertion1.event3, {'A', 'B'})

    def test_init_exceptions(self):
        self.assertRaises(ValueError, IndependenceAssertion, event2=['U'], event3='V')
        self.assertRaises(ValueError, IndependenceAssertion, event2=['U'])
        self.assertRaises(ValueError, IndependenceAssertion, event3=['Z'])
        self.assertRaises(ValueError, IndependenceAssertion, event1=['U'])
        self.assertRaises(ValueError, IndependenceAssertion, event1=['U'], event3=['Z'])

    def tearDown(self):
        del self.assertion


class TestIndependeciesAssertionEq(unittest.TestCase):
    def setUp(self):
        self.i1 = IndependenceAssertion('a', 'b', 'c')
        self.i2 = IndependenceAssertion('a', 'b')
        self.i3 = IndependenceAssertion('a', ['b', 'c', 'd'])
        self.i4 = IndependenceAssertion('a', ['b', 'c', 'd'], 'e')
        self.i5 = IndependenceAssertion('a', ['d', 'c', 'b'], 'e')
        self.i6 = IndependenceAssertion('a', ['d', 'c'], ['e', 'b'])
        self.i7 = IndependenceAssertion('a', ['c', 'd'], ['b', 'e'])
        self.i8 = IndependenceAssertion('a', ['f', 'd'], ['b', 'e'])
        self.i9 = IndependenceAssertion('a', ['d', 'k', 'b'], 'e')
        self.i10 = IndependenceAssertion(['k', 'b', 'd'], 'a', 'e')

    def test_eq1(self):
        self.assertFalse(self.i1 == 'a')
        self.assertFalse(self.i2 == 1)
        self.assertFalse(self.i4 == [2, 'a'])
        self.assertFalse(self.i6 == 'c')

    def test_eq2(self):
        self.assertFalse(self.i1 == self.i2)
        self.assertFalse(self.i1 == self.i3)
        self.assertFalse(self.i2 == self.i4)
        self.assertFalse(self.i3 == self.i6)

    def test_eq3(self):
        self.assertTrue(self.i4 == self.i5)
        self.assertTrue(self.i6 == self.i7)
        self.assertFalse(self.i7 == self.i8)
        self.assertFalse(self.i4 == self.i9)
        self.assertFalse(self.i5 == self.i9)
        self.assertTrue(self.i10 == self.i9)
        self.assertTrue(self.i10 != self.i8)

    def tearDown(self):
        del self.i1
        del self.i2
        del self.i3
        del self.i4
        del self.i5
        del self.i6
        del self.i7
        del self.i8
        del self.i9
        del self.i10


class TestIndependencies(unittest.TestCase):
    def setUp(self):
        self.Independencies = Independencies()
        self.Independencies3 = Independencies(['a', ['b', 'c', 'd'], ['e', 'f', 'g']],
                                              ['c', ['d', 'e', 'f'], ['g', 'h']])
        self.Independencies4 = Independencies([['f', 'd', 'e'], 'c', ['h', 'g']],
                                              [['b', 'c', 'd'], 'a', ['f', 'g', 'e']])
        self.Independencies5 = Independencies(['a', ['b', 'c', 'd'], ['e', 'f', 'g']],
                                              ['c', ['d', 'e', 'f'], 'g'])

    def test_init(self):
        self.Independencies1 = Independencies(['X', 'Y', 'Z'])
        self.assertEqual(self.Independencies1, Independencies(['X', 'Y', 'Z']))
        self.Independencies2 = Independencies()
        self.assertEqual(self.Independencies2, Independencies())

    def test_add_assertions(self):
        self.Independencies1 = Independencies(['X', 'Y', 'Z'])
        self.assertEqual(self.Independencies1, Independencies(['X', 'Y', 'Z']))
        self.Independencies2 = Independencies(['A', 'B', 'C'], ['D', 'E', 'F'])
        self.assertEqual(self.Independencies2, Independencies(['A', 'B', 'C'], ['D', 'E', 'F']))

    def test_get_assertions(self):
        self.Independencies1 = Independencies(['X', 'Y', 'Z'])
        self.assertEqual(self.Independencies1.independencies, self.Independencies1.get_assertions())
        self.Independencies2 = Independencies(['A', 'B', 'C'], ['D', 'E', 'F'])
        self.assertEqual(self.Independencies2.independencies, self.Independencies2.get_assertions())

    def test_closure(self):
        ind1 = Independencies(('A', ['B', 'C'], 'D'))
        self.assertEqual(ind1.closure(), Independencies(('A', ['B', 'C'], 'D'),
                                                        ('A', 'B', ['C', 'D']),
                                                        ('A', 'C', ['B', 'D']),
                                                        ('A', 'B', 'D'),
                                                        ('A', 'C', 'D')))
        ind2 = Independencies(('W', ['X', 'Y', 'Z']))
        self.assertEqual(ind2.closure(),
                         Independencies(
                             ('W', 'Y'), ('W', 'Y', 'X'), ('W', 'Y', 'Z'), ('W', 'Y', ['X', 'Z']),
                             ('W', ['Y', 'X']), ('W', 'X', ['Y', 'Z']), ('W', ['X', 'Z'], 'Y'),
                             ('W', 'X'), ('W', ['X', 'Z']), ('W', ['Y', 'Z'], 'X'),
                             ('W', ['Y', 'X', 'Z']), ('W', 'X', 'Z'), ('W', ['Y', 'Z']),
                             ('W', 'Z', 'X'), ('W', 'Z'), ('W', ['Y', 'X'], 'Z'), ('W', 'X', 'Y'),
                             ('W', 'Z', ['Y', 'X']), ('W', 'Z', 'Y')))
        ind3 = Independencies(('c', 'a', ['b', 'e', 'd']), (['e', 'c'], 'b', ['a', 'd']), (['b', 'd'], 'e', 'a'),
                              ('e', ['b', 'd'], 'c'), ('e', ['b', 'c'], 'd'), (['e', 'c'], 'a', 'b'))
        self.assertEqual(len(ind3.closure().get_assertions()), 78)

    def test_entails(self):
        ind1 = Independencies([['A', 'B'], ['C', 'D'], 'E'])
        ind2 = Independencies(['A', 'C', 'E'])
        self.assertTrue(ind1.entails(ind2))
        self.assertFalse(ind2.entails(ind1))
        ind3 = Independencies(('W', ['X', 'Y', 'Z']))
        self.assertTrue(ind3.entails(ind3.closure()))
        self.assertTrue(ind3.closure().entails(ind3))

    def test_is_equivalent(self):
        ind1 = Independencies(['X', ['Y', 'W'], 'Z'])
        ind2 = Independencies(['X', 'Y', 'Z'], ['X', 'W', 'Z'])
        ind3 = Independencies(['X', 'Y', 'Z'], ['X', 'W', 'Z'], ['X', 'Y', ['W', 'Z']])
        self.assertFalse(ind1.is_equivalent(ind2))
        self.assertTrue(ind1.is_equivalent(ind3))

    def test_eq(self):
        self.assertTrue(self.Independencies3 == self.Independencies4)
        self.assertFalse(self.Independencies3 != self.Independencies4)
        self.assertTrue(self.Independencies3 != self.Independencies5)
        self.assertFalse(self.Independencies4 == self.Independencies5)
        self.assertFalse(Independencies() == Independencies(['A', 'B', 'C']))
        self.assertFalse(Independencies(['A', 'B', 'C']) == Independencies())
        self.assertTrue(Independencies() == Independencies())

    def tearDown(self):
        del self.Independencies
        del self.Independencies3
        del self.Independencies4
        del self.Independencies5


if __name__ == '__main__':
    unittest.main()



import os
import unittest
import warnings
import numpy as np
import numpy.testing as np_test

from pgmpy.readwrite import XMLBIFReader, XMLBIFWriter
from pgmpy.models import BayesianModel
from pgmpy.factors.discrete import TabularCPD


try:
    from lxml import etree
except ImportError:
    try:
        import xml.etree.cElementTree as etree
    except ImportError:
        try:
            import xml.etree.ElementTree as etree
        except ImportError:
            warnings.warn("Failed to import ElementTree from any known place")

TEST_FILE = """<?xml version="1.0" encoding="US-ASCII"?>


<!--
    Bayesian network in XMLBIF v0.3 (BayesNet Interchange Format)
    Produced by JavaBayes (http://www.cs.cmu.edu/~javabayes/
    Output created Mon Aug 01 10:33:28 AEST 2016
-->



<!-- DTD for the XMLBIF 0.3 format -->
<!DOCTYPE BIF [
    <!ELEMENT BIF ( NETWORK )*>
          <!ATTLIST BIF VERSION CDATA #REQUIRED>
    <!ELEMENT NETWORK ( NAME, ( PROPERTY | VARIABLE | DEFINITION )* )>
    <!ELEMENT NAME (#PCDATA)>
    <!ELEMENT VARIABLE ( NAME, ( OUTCOME |  PROPERTY )* ) >
          <!ATTLIST VARIABLE TYPE (nature|decision|utility) "nature">
    <!ELEMENT OUTCOME (#PCDATA)>
    <!ELEMENT DEFINITION ( FOR | GIVEN | TABLE | PROPERTY )* >
    <!ELEMENT FOR (#PCDATA)>
    <!ELEMENT GIVEN (#PCDATA)>
    <!ELEMENT TABLE (#PCDATA)>
    <!ELEMENT PROPERTY (#PCDATA)>
]>


<BIF VERSION="0.3">
<NETWORK>
<NAME>Dog_Problem</NAME>

<!-- Variables -->
<VARIABLE TYPE="nature">
    <NAME>light_on</NAME>
    <OUTCOME>true</OUTCOME>
    <OUTCOME>false</OUTCOME>
    <PROPERTY>position = (73, 165)</PROPERTY>
</VARIABLE>

<VARIABLE TYPE="nature">
    <NAME>bowel_problem</NAME>
    <OUTCOME>true</OUTCOME>
    <OUTCOME>false</OUTCOME>
    <PROPERTY>position = (190, 69)</PROPERTY>
</VARIABLE>

<VARIABLE TYPE="nature">
    <NAME>dog_out</NAME>
    <OUTCOME>true</OUTCOME>
    <OUTCOME>false</OUTCOME>
    <PROPERTY>position = (155, 165)</PROPERTY>
</VARIABLE>

<VARIABLE TYPE="nature">
    <NAME>hear_bark</NAME>
    <OUTCOME>true</OUTCOME>
    <OUTCOME>false</OUTCOME>
    <PROPERTY>position = (154, 241)</PROPERTY>
</VARIABLE>

<VARIABLE TYPE="nature">
    <NAME>family_out</NAME>
    <OUTCOME>true</OUTCOME>
    <OUTCOME>false</OUTCOME>
    <PROPERTY>position = (112, 69)</PROPERTY>
</VARIABLE>


<!-- Probability distributions -->
<DEFINITION>
    <FOR>light_on</FOR>
    <GIVEN>family_out</GIVEN>
    <TABLE>0.6 0.4 0.05 0.95 </TABLE>
</DEFINITION>

<DEFINITION>
    <FOR>bowel_problem</FOR>
    <TABLE>0.01 0.99 </TABLE>
</DEFINITION>

<DEFINITION>
    <FOR>dog_out</FOR>
    <GIVEN>bowel_problem</GIVEN>
    <GIVEN>family_out</GIVEN>
    <TABLE>0.99 0.01 0.97 0.03 0.9 0.1 0.3 0.7 </TABLE>
</DEFINITION>

<DEFINITION>
    <FOR>hear_bark</FOR>
    <GIVEN>dog_out</GIVEN>
    <TABLE>0.7 0.3 0.01 0.99 </TABLE>
</DEFINITION>

<DEFINITION>
    <FOR>family_out</FOR>
    <TABLE>0.15 0.85 </TABLE>
</DEFINITION>


</NETWORK>
</BIF>"""


class TestXMLBIFReaderMethods(unittest.TestCase):

    def setUp(self):
        self.reader = XMLBIFReader(string=TEST_FILE)

    def test_get_variables(self):
        var_expected = ['light_on', 'bowel_problem', 'dog_out',
                        'hear_bark', 'family_out']
        self.assertListEqual(self.reader.variables, var_expected)

    def test_get_states(self):
        states_expected = {'bowel_problem': ['true', 'false'],
                           'dog_out': ['true', 'false'],
                           'family_out': ['true', 'false'],
                           'hear_bark': ['true', 'false'],
                           'light_on': ['true', 'false']}
        states = self.reader.variable_states
        for variable in states_expected:
            self.assertListEqual(states_expected[variable],
                                 states[variable])

    def test_get_parents(self):
        parents_expected = {'bowel_problem': [],
                            'dog_out': ['bowel_problem', 'family_out'],
                            'family_out': [],
                            'hear_bark': ['dog_out'],
                            'light_on': ['family_out']}
        parents = self.reader.variable_parents
        for variable in parents_expected:
            self.assertListEqual(parents_expected[variable],
                                 parents[variable])

    def test_get_edges(self):
        edges_expected = [['family_out', 'dog_out'],
                          ['bowel_problem', 'dog_out'],
                          ['family_out', 'light_on'],
                          ['dog_out', 'hear_bark']]
        self.assertListEqual(sorted(self.reader.edge_list),
                             sorted(edges_expected))

    def test_get_values(self):
        cpd_expected = {'bowel_problem': np.array([[0.01],
                                                   [0.99]]),
                        'dog_out': np.array([[0.99, 0.97, 0.9, 0.3],
                                             [0.01, 0.03, 0.1, 0.7]]),
                        'family_out': np.array([[0.15],
                                                [0.85]]),
                        'hear_bark': np.array([[0.7, 0.01],
                                               [0.3, 0.99]]),
                        'light_on': np.array([[0.6, 0.05],
                                              [0.4, 0.95]])}
        cpd = self.reader.variable_CPD
        for variable in cpd_expected:
            np_test.assert_array_equal(cpd_expected[variable],
                                       cpd[variable])

    def test_get_property(self):
        property_expected = {'bowel_problem': ['position = (190, 69)'],
                             'dog_out': ['position = (155, 165)'],
                             'family_out': ['position = (112, 69)'],
                             'hear_bark': ['position = (154, 241)'],
                             'light_on': ['position = (73, 165)']}
        prop = self.reader.variable_property
        for variable in property_expected:
            self.assertListEqual(property_expected[variable],
                                 prop[variable])

    def test_model(self):
        self.reader.get_model().check_model()

    def tearDown(self):
        del self.reader


class TestXMLBIFReaderMethodsFile(unittest.TestCase):

    def setUp(self):
        with open("dog_problem.xml", 'w') as fout:
            fout.write(TEST_FILE)
        self.reader = XMLBIFReader("dog_problem.xml")

    def test_get_variables(self):
        var_expected = ['light_on', 'bowel_problem', 'dog_out',
                        'hear_bark', 'family_out']
        self.assertListEqual(self.reader.variables, var_expected)

    def test_get_states(self):
        states_expected = {'bowel_problem': ['true', 'false'],
                           'dog_out': ['true', 'false'],
                           'family_out': ['true', 'false'],
                           'hear_bark': ['true', 'false'],
                           'light_on': ['true', 'false']}
        states = self.reader.variable_states
        for variable in states_expected:
            self.assertListEqual(states_expected[variable],
                                 states[variable])

    def test_get_parents(self):
        parents_expected = {'bowel_problem': [],
                            'dog_out': ['bowel_problem', 'family_out'],
                            'family_out': [],
                            'hear_bark': ['dog_out'],
                            'light_on': ['family_out']}
        parents = self.reader.variable_parents
        for variable in parents_expected:
            self.assertListEqual(parents_expected[variable],
                                 parents[variable])

    def test_get_edges(self):
        edges_expected = [['family_out', 'dog_out'],
                          ['bowel_problem', 'dog_out'],
                          ['family_out', 'light_on'],
                          ['dog_out', 'hear_bark']]
        self.assertListEqual(sorted(self.reader.edge_list),
                             sorted(edges_expected))

    def test_get_values(self):
        cpd_expected = {'bowel_problem': np.array([[0.01],
                                                   [0.99]]),
                        'dog_out': np.array([[0.99, 0.97, 0.9, 0.3],
                                             [0.01, 0.03, 0.1, 0.7]]),
                        'family_out': np.array([[0.15],
                                                [0.85]]),
                        'hear_bark': np.array([[0.7, 0.01],
                                               [0.3, 0.99]]),
                        'light_on': np.array([[0.6, 0.05],
                                              [0.4, 0.95]])}
        cpd = self.reader.variable_CPD
        for variable in cpd_expected:
            np_test.assert_array_equal(cpd_expected[variable],
                                       cpd[variable])

    def test_get_property(self):
        property_expected = {'bowel_problem': ['position = (190, 69)'],
                             'dog_out': ['position = (155, 165)'],
                             'family_out': ['position = (112, 69)'],
                             'hear_bark': ['position = (154, 241)'],
                             'light_on': ['position = (73, 165)']}
        prop = self.reader.variable_property
        for variable in property_expected:
            self.assertListEqual(property_expected[variable],
                                 prop[variable])

    def test_model(self):
        self.reader.get_model().check_model()

    def tearDown(self):
        del self.reader
        os.remove("dog_problem.xml")


class TestXMLBIFWriterMethodsString(unittest.TestCase):

    def setUp(self):
        reader = XMLBIFReader(string=TEST_FILE)
        self.expected_model = reader.get_model()
        self.writer = XMLBIFWriter(self.expected_model)

        self.model_stateless = BayesianModel([('D', 'G'), ('I', 'G'), ('G', 'L'), ('I', 'S')])
        self.cpd_d = TabularCPD(variable='D', variable_card=2, values=[[0.6, 0.4]])
        self.cpd_i = TabularCPD(variable='I', variable_card=2, values=[[0.7, 0.3]])

        self.cpd_g = TabularCPD(variable='G', variable_card=3,
                                values=[[0.3, 0.05, 0.9,  0.5],
                                        [0.4, 0.25, 0.08, 0.3],
                                        [0.3, 0.7,  0.02, 0.2]],
                                evidence=['I', 'D'],
                                evidence_card=[2, 2])

        self.cpd_l = TabularCPD(variable='L', variable_card=2,
                                values=[[0.1, 0.4, 0.99],
                                        [0.9, 0.6, 0.01]],
                                evidence=['G'],
                                evidence_card=[3])

        self.cpd_s = TabularCPD(variable='S', variable_card=2,
                                values=[[0.95, 0.2],
                                        [0.05, 0.8]],
                                evidence=['I'],
                                evidence_card=[2])

        self.model_stateless.add_cpds(self.cpd_d, self.cpd_i, self.cpd_g, self.cpd_l, self.cpd_s)
        self.writer_stateless = XMLBIFWriter(self.model_stateless)

    def test_write_xmlbif_statefull(self):
        self.writer.write_xmlbif("dog_problem_output.xbif")
        with open("dog_problem_output.xbif", "r") as f:
            file_text = f.read()
        reader = XMLBIFReader(string=file_text)
        model = reader.get_model()
        self.assert_models_equivelent(self.expected_model, model)
        os.remove("dog_problem_output.xbif")

    def test_write_xmlbif_stateless(self):
        self.writer_stateless.write_xmlbif("grade_problem_output.xbif")
        with open("grade_problem_output.xbif", 'r') as f:
            reader = XMLBIFReader(f)
        model = reader.get_model()
        self.assert_models_equivelent(self.model_stateless, model)
        self.assertDictEqual({'G': ['state0', 'state1', 'state2'],
                              'I': ['state0', 'state1'],
                              'D': ['state0', 'state1'],
                              'S': ['state0', 'state1'],
                              'L': ['state0', 'state1']},
                             model.get_cpds('D').state_names)
        os.remove("grade_problem_output.xbif")

    def assert_models_equivelent(self, expected, got):
        self.assertSetEqual(set(expected.nodes()), set(got.nodes()))
        for node in expected.nodes():
            self.assertListEqual(expected.get_parents(node), got.get_parents(node))
            cpds_expected = expected.get_cpds(node=node)
            cpds_got = got.get_cpds(node=node)
            np_test.assert_array_equal(cpds_expected.values, cpds_got.values)
